/* program to do a dithered rescale from floating point file to integer file.
   It will create an integer file called  filename.snd if you don't
   specify an output file. */

/* TODO create option to skip on input and output, and specify duration and 
   optional peak */

#include "rescale.h"
#include "../H/sfheader.h"
#include <stdio.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <errno.h>
#include <bsd_mem.h>

#define  BUFSIZE 32768

extern	char * getsfcode(SFHEADER *,int);
extern putsfcode(SFHEADER *,void *,SFCODE *);
extern writeHeader(int,SFHEADER  *);	/* writes out whatever type of header was read in */
extern int putlength(char *sfname,int sf,SFHEADER *header);
extern void printsf(SFHEADER *);
extern 	ftruncate(int,long long);
extern 	close(int);

float drandom(long *);

static SFCODE	ampcode = {
	SF_MAXAMP,
	sizeof(SFMAXAMP) + sizeof(SFCODE)
}; 

static SFCODE	commentcode = {
	SF_COMMENT,
	MINCOMM + sizeof(SFCODE)
	};

/* adapted version of rrand, returns values between -1.0 and 1.0 which are */
/* the sum of two random number generators */

float drandom(long *seeds)
{
	int i1 = ((*seeds = *seeds * 1103515245 + 12345)>>16) & 077777;
	int i2 = ((*(seeds+1) = *(seeds+1) * 1103515245 + 12345)>>16) & 077777;
	return ((float)i1/32768. + (float)i2/32768. - 1.0);
}

void
main(int argc,char **argv)
{
	SFMAXAMP sfm,sfmnew;
	SFCOMMENT sfcm;
	SFHEADER sfh1,sfh2;
	SFCODE *sizer;
	struct stat sfst1,sfst2;

	char *cp,*sfin,*sfout,*getsfcode();
	double atof();
//	short outbuffer[BUFSIZE], *outptr;
//	float inbuffer[BUFSIZE], *inptr;
	short *outbuffer, *outptr;
	float *inbuffer, *inptr;
//	FILE * sf1;
	int i,bytes,words,inbytes,outbytes,sf1,sf2,result,readbytes,durbytes;
	int replace,empty,skipbytes,outskipbytes,newfile;
	float opeak,factor,skpin,dur,newpeak,outskip;
//	char command[256];
//	char newname[128];
	char *point,*strcat();
	short *bufp;
	float Oresult = 32767;
	float filedur;
	long seeds[2];
	int headersize = SIZEOF_BSD_HEADER;

	if(argc == 1) {
usage: printf("usage:  -s inskip -d dur -o outskip -p peakamp -P desired peak\n");
printf ("\t-r  [write over floating-point file] -e\n");
printf ("\temptybuffers_at_end -z [create new file] -f factor\n");
printf ("\tinputfile [outputfile]\n");
		printf("defaults: s=0, d=to_eof, o=0, p=input_peak, e=1, P=32767,\n");
		printf ("\tf=P/input_peak write over old file, outputfile=inputfile.snd\n");
		exit(0);
		}
	replace=newfile=skpin=dur=newpeak=outskip=factor=0;

	empty = 1;  /* defaults to 1 empty buffers at end*/

	while((*++argv)[0] == '-') {
		argc -= 2; /* Take away two args */
		for(cp = argv[0]+1; *cp; cp++) {
			switch(*cp) { /* Grap options */
			case 'r':
				replace = 1;
				printf("Writing over floating-point file\n");
				break;
			case 's': 
				skpin = atof(*++argv);
				printf("input skip = %f\n",skpin);
				break;
			case 'd':
				dur = atof(*++argv);
				printf("rescale duration = %f\n",dur);
				empty = 0; /* no empty buffers for splice */
				break;
			case 'f':
				factor = atof(*++argv);
				printf("specified factor = %f\n",factor);
				break;
			case 'p': 
				newpeak = atof(*++argv);
				printf("specified peak = %f\n",newpeak);
				break;
			case 'P':
				Oresult = atof(*++argv);
				printf("resultant peak = %f\n",Oresult);
				break;
			case 'o': 
				outskip = atof(*++argv);
				printf("output skip =  %f\n",outskip);
				break;
			case 'e':
				empty = atof(*++argv);
				printf("write %d empty buffers at end\n",empty);
				break;
			case 'z':
				newfile = 1;
				break;
			default:
				printf("uh-oh, unkown option\n");
				goto usage;
			}
		}
	}
	sfin = argv[0];
	sfout = argv[1];
//	readopensf(sfin,sf1,sfh1,sfst1,"rescale",result);
	newrwopensf(sfin,sf1,sfh1,sfst1,"rescale",result,0);
	if(result < 0) {
		close(sf1);
		exit(1);
	}
//
// print duration of file
//
	filedur = (float)(sfst1.st_size - headersize)
		/(float)sfclass(&sfh1)/(float)sfchans(&sfh1)
		/sfsrate(&sfh1);
	printf ("input file duration: %f seconds\n",filedur);

/* do input skip on input file*/
	if(skipbytes = 

		skpin * sfclass(&sfh1) * sfsrate(&sfh1) * sfchans(&sfh1)) {
		skipbytes -= skipbytes % (sfclass(&sfh1) * sfchans(&sfh1));
			/* make sure it lands on sample block */
		if(sflseek(sf1,skipbytes,0) == -1) {
			printf("Bad skip on input file\n");
			exit(1);
		}
	}

	printsf(&sfh1);
	if(sfclass(&sfh1) == SF_SHORT) {
		printf("Note: Input file has short integers.\n");
	}

	cp = getsfcode(&sfh1,SF_MAXAMP);
	bcopy(cp + sizeof(SFCODE), (char *) &sfm, sizeof(SFMAXAMP));
	for(i=0,opeak=0; i<sfchans(&sfh1); i++) 

		if(sfmaxamp(&sfm,i) > opeak) opeak = sfmaxamp(&sfm,i);
	opeak = newpeak ? newpeak : opeak;
	if(!opeak) {
		printf("Sorry, but I have no peak amplitude for this file.\n");
		printf("Put one in with sfhedit, sndpeak, or use the -p flag in rescale.\n");
		close(sf1);
		exit(-3);
		}
	printf("Peak amplitude of input file is %e\n",opeak);
	if(!factor) factor = Oresult/opeak;
	printf("factor = %f\n",factor);

	if((cp=getsfcode(&sfh1,SF_COMMENT)))  {
	sizer = (SFCODE *) cp;
	bcopy(cp + sizeof(SFCODE) , (char *) &sfcm, sizer->bsize);
	}
	if(replace) sfout=sfin;
	/* added to allow compatibility with new i/o -- DAS */
#if defined(NeXT) || defined(NEXT)
	sfh2.sfinfo.NeXTheader.magic = SND_MAGIC;
	sfmagic(&sfh2) = SF_MAGIC;
#endif
	newrwopensf(sfout,sf2,sfh2,sfst2,"rescale",result,2);
	if(result < 0) {
		if(sfout == NULL) {
			point = sfin;
			if(sfclass(&sfh1) == SF_FLOAT) 

				point = strcat(sfin,".snd");
			else
				point = strcat(sfin,".xshort.snd");
		}
		else
			point = sfout;

		sfmagic(&sfh2) = SF_MAGIC;
		sfclass(&sfh2) = SF_SHORT;
		sfchans(&sfh2) = sfchans(&sfh1);
		sfsrate(&sfh2) = sfsrate(&sfh1);
		if((sf2 = open(point,O_CREAT|O_RDWR,0644)) < 0 ) {
			printf("Can't open file %s\n",point);
			exit(-2);
		}
		if(newfile) ftruncate(sf2,SIZEOF_NeXT_HEADER);

		for(i=0; i<sfchans(&sfh2); i++) {
			sfmaxamp(&sfmnew,i)=sfmaxamp(&sfm,i)*factor;
			sfmaxamploc(&sfmnew,i)=sfmaxamploc(&sfm,i);
		}
		sfmaxamptime(&sfmnew) = sfmaxamptime(&sfm);
		putsfcode(&sfh2,&sfmnew,&ampcode);
		if (putsfcode(&sfh2,&sfcm,&commentcode) < 0) {
			printf("comment didn't get written, sorry!\n");
			exit(-1);
		}
		printf
			("\nCreating output file: %s\n",point);
		if(wheader(sf2,&sfh2)) {
			printf("Can't seem to write header on file %s\n", point);
			perror("wheader");
			exit(-1);
		}
	}
	else if(!replace) printsf(&sfh2);
	if(!replace && (sfclass(&sfh2) != SF_SHORT)) {
		printf("Output file must have short integers.\n");
		exit(-1);
	}
	/* do output skip*/
	if(outskipbytes = 

		  outskip * sfclass(&sfh2) * sfsrate(&sfh2) * sfchans(&sfh2)) {
		outskipbytes -= 

			      outskipbytes % (sfclass(&sfh2) * sfchans(&sfh2));
			/* make sure it lands on sample block */
		if(sflseek(sf2,outskipbytes,0) == -1) {
			printf("Bad skip on output file\n");
			exit(1);
		}
	}

	readbytes = inbytes = BUFSIZE * sfclass(&sfh1);

	durbytes = dur * sfclass(&sfh1) * sfchans(&sfh1) * sfsrate(&sfh1);

	fflush (stdout);
	inbuffer = (float *)calloc(BUFSIZE, sizeof(float));
	outbuffer = (short *)calloc(BUFSIZE, sizeof(short));
	bufp = (short *)inbuffer;

	fprintf(stderr,"Rescaling....\t");

	seeds[0] = 17;
	seeds[1] = 31;		/* seeds for dither */

	while(1) {
		if(dur) {
			inbytes = (durbytes > readbytes) ? inbytes : durbytes;
			durbytes -= inbytes;
		}
		if((bytes = read(sf1,(char *)inbuffer,inbytes)) <= 0) {
			printf("reached eof on input\n");
			close(sf1);
			break;
		}
		words = bytes/sfclass(&sfh1);
//printf("bytes %d words %d class %d\n", bytes, words, sfclass(&sfh1));
		outbytes = words * SF_SHORT;
		inptr = inbuffer;
		outptr = outbuffer;
		if(sfclass(&sfh1) == SF_SHORT)
			for(i=0; i<words; i++) 

				*outptr++ = (short)((float)bufp[i] * factor);
		else
			for(i=0; i<words; i++)
				*outptr++ = (short) ((*inptr++ * factor) + drandom(seeds));

		if(write(sf2,(char *)outbuffer,outbytes) != outbytes) {
			printf("Bad write on output file\n");
			free(inbuffer);
			free(outbuffer);
			close(sf1);
			close(sf2);
			exit(0);
		}
	}

//		if (fwrite((char *)outbuffer, sizeof(short), outbytes, sf2) != outbytes 

/* write  empty buffers
	commented out because it writes too much 0 data at eof
*/
//	for(i=0; i<words; i++) outbuffer[i] = 0;
//	for(i=0; i<empty; i++)
//		if(n=write(sf2,(char *)outbuffer,outbytes) != outbytes) 
//			printf("Bad write on output file\n");
//			close(sf2);
//			exit(0);
//		
	if(replace) {
		i = lseek(sf2,0,1);
		if(ftruncate(sf2,i) < 0) 

			printf("Bad truncation\n");
		lseek(sf2,0,0);
		for(i=0; i<sfchans(&sfh2); i++) {
			sfmaxamp(&sfmnew,i)=sfmaxamp(&sfm,i)*factor;
			sfmaxamploc(&sfmnew,i)=sfmaxamploc(&sfm,i);
		}
		sfmaxamptime(&sfmnew) = sfmaxamptime(&sfm);
		putsfcode(&sfh2,&sfmnew,&ampcode);
		sfclass(&sfh2) = SF_SHORT;
		if(wheader(sf2,&sfh2)) {
			printf("Can't seem to write header on file %s\n", point);
			perror("main");
			exit(-1);
		}
//		if(fsync(sf2) < 0) printf("bad fsync\n");
	}
#if defined(NeXT) || defined(NEXT)
	putlength(point,sf2,&sfh2);
#endif
	free(inbuffer);
	free(outbuffer);
	close(sf1); close(sf2);
	printf("\ndone.\n");
}
