#include "../H/sfheader.h"
#include "../H/ugens.h"
#include "../macros/macros.h"
#include <stdio.h>
#include <unistd.h>
#include <bsd_mem.h>

#define MAXSIZE 32768

//int NBYTES = 32768;

float *q;

extern hplset(float xxlp,float dur,float dynam,float plamp,float seed,float sr,int init,float *q);
extern void evset(float ,float ,float ,int ,float *);
extern void combset(float ,float ,int ,float *);
extern double evp(int ,float *, float *, float *);
extern double comb(float ,float *);
extern int setnote(double ,double ,int);
extern double cpspch(float pch);

float frand(float x);
double pluckit(float *p,int n_args, double *pp);
double hpluckit(float *p,int n_args, double *pp);
double allpassit(float *p,int n_args, double *pp);
double combit(float *p,int n_args, double *pp);
double pluckitin(float *p,int n_args, double *pp);
double hpluckitin(float *p,int n_args, double *pp);
void profile(void);

double
pluckit(float *p,int n_args, double *pp)
{
/* simple routine to do plucked string filter
pp[0]=start,p1=dur, p2=pitch, p3=null [4] = relamppluck, p5=rise
p6 = decay p7=null p8=null p9=seed*/

	int output,nsamps,j,i,control;
	float dynam,loopt,pluckamp,seed;
	float evals[9],*f1,out[4],enval; 
	float start,dur;
	int len;
	
	if(n_args < 10) {
		printf("start,dur,pitch,null,relamppluck,rise,decay,null,coef,seed\n");
		return;
		}
	output=1;
	start = pp[0];
	dur = pp[0]+pp[1] - start;

	nsamps = setnote(start,dur,output);

	loopt = 1./cpspch(pp[2]);
	f1 = (float *)floc(1);
	pluckamp = pp[4];
	seed = pp[9];
	dynam=pp[8];
	len = loopt * SR + 20;
	printf("len = %d\n",len);
	if((q = (float *)malloc((int)(len *  4))) == 0) {
		printf("can't malloc memory for delayline\n");
		exit(-1);
	}
	pluckset(loopt,pluckamp,seed,dynam,q,SR);

	evset(dur,pp[5],pp[6],1,evals);
	control = SR/cpspch(pp[2]);

	for(j=0,i=0; i<nsamps; i++) {
		if(!j--) {
			enval = evp(i,f1,f1,evals);
			j=control;
		}

		out[0] = pluck(0.,q);

		ADDOUT(out,output);
	}
	endnote(output);
	free(q);
}

double
hpluckit(float *p,int n_args, double *pp)
{
/* simple routine to do plucked string filter
pp[0]=start,p1=dur, p2=pitch, p3=null [4] = relamppluck, p5=rise
p6 = decay p7=realdurpluck p8=brightness p9=seed*/

	int output,nsamps,j,i,control;
	float dynam,loopt,realpluckdur,pluckamp,seed;
	float evals[9],*f1,out[4],enval; 
	float start,dur;
	int len;

	if(n_args < 10) {
	printf("start,dur,pitch,null,relamppluck,rise,decay,realdurpluck,brightness,seed\n");
		return;
	}

	output=1;
	start = pp[0];
	dur = pp[0]+pp[1] - start;

	nsamps = setnote(start,dur,output);

	loopt = 1./cpspch(pp[2]);
	realpluckdur = (n_args == 8) ? pp[7] : pp[1];
	f1 = (float *)floc(1);
	pluckamp = pp[4];
	seed = pp[9];
	dynam=pp[8];

	len = loopt * SR + 20;
	printf("len = %d\n",len);
	if((q = (float *)malloc((int)(len *  4))) == 0) {
		printf("can't malloc memory for delayline\n");
		exit(-1);
	}
	hplset(loopt,realpluckdur,dynam,pluckamp,seed,SR,1,q);

	evset(dur,pp[5],pp[6],1,evals);

	control = SR/cpspch(pp[2]);

	for(j=0,i=0; i<nsamps; i++) {
		if(!j--) {
			enval = evp(i,f1,f1,evals);
			j=control;
		}

		HPLUCK(out[0],0.,q);

		out[0] *= enval;

		ADDOUT(out,output);
	}
	endnote(output);
	free(q);
}

double
allpassit(float *p,int n_args, double *pp)
{
/*pp[0]=start,p1=dur, p2=pitch, p3=inskip [4] = rvt, p5=rise
p6 = decay p7=null p8=null p9=null*/

	int input,output,nsamps,j,i,control;
	float rvt,loopt;
	float evals[9],*f1,xin[4],out[4],enval; 
	float hz;
	float start,dur,inskip;
	int len,oldnsamp;
	if(n_args < 7) {
		printf("start, dur, pitch, inskip, rvt, rise, decay\n");
		return;
		}
	input=0; output=1;
	start = pp[0];
	dur = pp[0]+pp[1] - start;
	inskip = pp[3];
	printf("start=%f, dur=%f inskip=%f\n",start,dur,inskip);

	nsamps = setnote(start,dur,output);
		setnote(inskip,dur,input);

	hz = (pp[2] > 0) ? cpspch(pp[2]) : -pp[2];
	loopt = 1./hz;
	f1 = (float *)floc(1);
	rvt=pp[4];

	len = loopt * SR + 20;
	printf("len = %d\n",len);
	if((q = (float *)malloc((int)(len *  4))) == 0) {
		printf("can't malloc memory for delayline\n");
		exit(-1);
	}
	combset(loopt,rvt,0,q);
	evset(dur+rvt,pp[5],pp[6],1,evals);

	control = SR/hz;

	for(j=0,i=0; i<nsamps; i++) {
		if(!j--) {
			enval = evp(i,f1,f1,evals);
			j=control;
		}
		if(!GETIN(xin,input)) break;

		out[0] = allpass(xin[0],q);

		out[0] *= enval;

		ADDOUT(out,output);
	}
	oldnsamp = nsamps;
	nsamps=rvt*SR;
	for(j=0,i=0; i<nsamps; i++) {
		if(!j--) {
			enval = evp(i+oldnsamp,f1,f1,evals);
			j=control;
		}
		out[0] = allpass(0.,q);
		out[0] *= enval;
		ADDOUT(out,output);
	}
	endnote(output);
	free(q);

}

double
combit(float *p,int n_args, double *pp)
{
/*pp[0]=start,p1=dur, p2=pitch, p3=inskip [4] = rvt, p5=rise
p6 = decay p7=null p8=null p9=null*/

	int input,output,nsamps,j,i,control;
	float rvt,loopt;
	float evals[9],*f1,xin[4],out[4],enval; 
	double hz=0;
	double start,dur,inskip;
	int len,oldnsamp;
	if(n_args < 7) {
		printf("start, dur, pitch, inskip, rvt, rise, decay\n");
		return;
		}
	input=0; output=1;
	start = pp[0];
	dur = pp[0]+pp[1] - start;
	inskip = pp[3];
	printf("start=%f, dur=%f inskip=%f ",start,dur,inskip);

	nsamps = setnote(start,dur,output);
		setnote(inskip,dur,input);

	hz = (pp[2] > 0) ? cpspch(pp[2]) : -pp[2];
	printf("hz = %f\n",hz);
	loopt = 1./hz;
	f1 = (float *)floc(1);
	rvt=pp[4];

	len = loopt * SR + 20;
	printf("len = %d\n",len);
	if((q = (float *)malloc((int)(len *  4))) == 0) {
		printf("can't malloc memory for delayline\n");
		exit(-1);
	}

	combset(loopt,rvt,0,q);
	evset(dur+rvt,pp[5],pp[6],1,evals);

	control = SR/hz;

	for(j=0,i=0; i<nsamps; i++) {
		if(!j--) {
			enval = evp(i,f1,f1,evals);
			j=control;
		}
		if(!GETIN(xin,input)) break;

		out[0] = comb(xin[0],q);

		out[0] *= enval;

		ADDOUT(out,output);
	}
	oldnsamp = nsamps;
	nsamps=rvt*SR;

	for(j=0,i=0; i<nsamps; i++) {
		if(!j--) {
			enval = evp(i+oldnsamp,f1,f1,evals);
			j=control;
		}
		out[0] = comb(0.,q);
		out[0] *= enval;
		ADDOUT(out,output);
	}
	endnote(output);
	free(q);
}

double
pluckitin(float *p,int n_args, double *pp)
{
/* simple routine to read input signal into plucked string filter
pp[0]=start,p1=dur, p2=pitch, p3=inskip [4] = relamppluck, p5=rise
p6 = decay p7=realdurpluck p8=brightness p9=seed*/

	int input,output,nsamps,j,i,control;
	float dynam,loopt,realpluckdur,pluckamp,seed;
	float evals[9],*f1,xin[4],out[4],enval; 
	float start,dur,inskip;
	int len;

	if(n_args < 10) {
		printf("start, dur, pitch, inskip, amp, rise, decay, null, coef, seed\n");
		return;
		}
	input=0; output=1;
	start = pp[0];
	dur = pp[0]+pp[1] - start;
	inskip = pp[3];
	printf("start=%f, dur=%f inskip=%f\n",start,dur,inskip);

	nsamps = setnote(start,dur,output);
		setnote(inskip,dur,input);

	loopt = 1./cpspch(pp[2]);
	realpluckdur = (n_args == 8) ? pp[7] : pp[1];
	f1 = (float *)floc(1);
	pluckamp = pp[4];
	seed = pp[9];
	dynam=pp[8];
	len = loopt * SR + 25;
	if((q = (float *)malloc((int)(len * 4))) == 0) {
		printf("can't malloc memory for delayline\n");
		exit(-1);
	}
	pluckset(loopt,pluckamp,seed,dynam,q,SR);

	evset(dur,pp[5],pp[6],1,evals);

	control = SR/cpspch(pp[2]);

	for(j=0,i=0; i<nsamps; i++) {
		if(!j--) {
			enval = evp(i,f1,f1,evals);
			j=control;
		}
		if(!GETIN(xin,input)) break;

		out[0] = pluck(xin[0],q);

		out[0] *= enval;

		ADDOUT(out,output);
	}
	endnote(output);
	free(q);
}
double
hpluckitin(float *p,int n_args, double *pp)
{
/* simple routine to read input signal into plucked string filter
pp[0]=start,p1=dur, p2=pitch, p3=inskip [4] = relamppluck, p5=rise
p6 = decay p7=realdurpluck p8=brightness p9=seed*/

	int input,output,nsamps,j,i,control;
	float dynam,loopt,realpluckdur,pluckamp,seed;
	float evals[9],*f1,xin[4],out[4],enval; 
	float start,dur,inskip;
	int len;

	if(n_args < 10) {
		printf("start, dur, pitch, inskip, relamppluck, rise, decay, realpluckdur, brightness, seed\n");
		return;
		}
	input=0; output=1;
	start = pp[0];
	dur = pp[0]+pp[1] - start;
	inskip = pp[3];
	printf("start=%f, dur=%f inskip=%f\n",start,dur,inskip);

	nsamps = setnote(start,dur,output);
		setnote(inskip,dur,input);

	loopt = 1./cpspch(pp[2]);
	realpluckdur = (n_args == 8) ? pp[7] : pp[1];
	f1 = (float *)floc(1);
	pluckamp = pp[4];
	seed = pp[9];
	dynam=pp[8];
	len = loopt * SR + 25;
	printf("len = %d\n",len);
	if((q = (float *)malloc((int)(len * 4))) == 0) {
		printf("can't malloc memory for delayline\n");
		exit(-1);
	}
	hplset(loopt,realpluckdur,dynam,pluckamp,seed,SR,1,q);

	evset(dur,pp[5],pp[6],1,evals);

	control = SR/cpspch(pp[2]);

	for(j=0,i=0; i<nsamps; i++) {
		if(!j--) {
			enval = evp(i,f1,f1,evals);
			j=control;
		}
		if(!GETIN(xin,input)) break;

		HPLUCK(out[0],xin[0],q);
		out[0] *= enval;

		ADDOUT(out,output);
	}
	endnote(output);
	free(q);
}

void
profile(void)
{
double pp[10];
float *p;
	p=0;
	UG_INTRO("pluckit",pluckit);
	UG_INTRO("hpluckit",hpluckit);
	UG_INTRO("combit",combit);
	UG_INTRO("allpassit",allpassit);
	UG_INTRO("pluckitin",pluckitin);
	UG_INTRO("hpluckitin",hpluckitin);
	pp[0]=1; pp[1]=7; pp[2]=512; pp[3]=0; pp[4]=512; pp[5]=1;
	makegen(p,6,pp);
}
