#include <stdlib.h>
#include <stdio.h>
#include "crack.h"
#include "pv.h"

complex zero = { 0., 0. };
complex one = { 1., 0. };
float PI;
float TWOPI;
float synt = 0.;

main(argc, argv)
    int argc; char *argv[];
{
    int 	R,
		N,
		N2,
		Nw = 0,
		Nw2, 
		D = 0, 
		I = 0,
		i,
		chanindex = 0,
		chanits,
		chanoffset,
		chanadd = 0,
		hoop,			/* looping variable */
		in,
		on,
		eof = 0,
		aflag = 0,
		sflag = 0,
		Np = 0;
    float 	P = 0.,
		len = 0.,
		tincr,
		tpos,
		*tvf,
		*Hwin,
		*Wanal,
		*Wsyn,
		*input,
		*winput,
		*lpcoef,
		*buffer,
		*channel,
		*output,
		lpa(),
		maxamp();
    char	ch,
		*dbuf;

    if (isatty(0))
	usage(1);

    while( (ch= crack( argc, argv, "R|N|M|D|I|P|p|l|g|ah", 0  )) != NULL ) {
	switch(ch) {
	    case 'R':	R = atoi(arg_option);
			break;
	    case 'N':	N = atoi(arg_option);
			break;
	    case 'M':	Nw = atoi(arg_option);
			break;
	    case 'D':	D = atoi(arg_option);
			break;
	    case 'I':	I = atoi(arg_option);
			break;
	    case 'P':	P = atof(arg_option);
			break;
	    case 'p':	Np = atoi(arg_option);
			break;
	    case 'g':	synt = atof(arg_option);
			break;
	    case 'l':	len = atof(arg_option);
			break;
	    case 'a':	aflag = 1;
			break;
	    case 's':	sflag = 1;
			break;
	    case 'h':	usage(1);
	}
    }

    if (len == 0.) {
      fprintf(stderr,"Specify length of analysis");
      usage(1);
    }

    if (Nw == 0)
	Nw = N;

    if (aflag && sflag) {
	fprintf(stderr,"specify either -a or -s not both\n");
	exit(1);
    }

    if (aflag)
	I = 0;
    else
	if (I == 0)
	    I = D;

    if (sflag)
	D = 0;

    PI = 4.*atan(1.);
    TWOPI = 8.*atan(1.);
    N2 = N>>1;
    Nw2 = Nw>>1;
    chanits = ((float) R / D) * len;

/* rewrite with NX malloc() routines */

    lpcoef = (float *) space( Np+1, sizeof(float) );	/* lpc coefficients */
    Wanal = (float *) space( Nw, sizeof(float) );	/* analysis window */
    Wsyn = (float *) space( Nw, sizeof(float) );	/* synthesis window */
    input = (float *) space( Nw, sizeof(float) );	/* input buffer */

    /* plain Hamming window */
    Hwin = (float *) space( Nw, sizeof(float) );	

    /* windowed input buffer */
    winput = (float *) space( Nw, sizeof(float) );	

    buffer = (float *) space( N, sizeof(float) );	/* FFT buffer */
    output = (float *) space( Nw, sizeof(float) );	/* output buffer */

    /* analysis channels */
    fprintf(stderr,"Allocating %d bytes for analysis array\n",
	(N+2) * ((int) chanits) * sizeof(float));
    channel = (float *) space( (N+2) * ((int) chanits),	sizeof(float) );

/* create windows */

    makewindows( Hwin, Wanal, Wsyn, Nw, N, I, 0 );

/* initialize input and output time values (in samples) */
    in = -Nw;
    chanoffset = (N+2);

    while ( !eof || chanindex < chanits ) {

/* increment times */
      in += D;

/* analysis: input D samples; window, fold and rotate input
   samples into FFT buffer; take FFT; and convert to
   amplitude-frequency (phase vocoder) form */

      if ( D == 0 ) {
	for ( hoop=0; hoop < N+2; hoop++ ) {
	  if ( fread(channel+hoop+chanadd,sizeof(float),1,stdin) <= 0 )
	    eof = 1;
	}
      } 
      else {
	eof = shiftin( input, Nw, D );
	fold( input, Wanal, Nw, buffer, N, in );
	rfft( buffer, N2, FORWARD );
	leanconvert( buffer, channel+chanadd, N2 );
      }
      ++chanindex;
      chanadd += chanoffset;
    }

    chanindex = 0;
    chanadd = 0;
    on = -Nw;
    chanoffset = (N+2);
    while ( chanindex < chanits ) {

      on += I;
      if ( aflag ) {  /* analysis output */
	fwrite( channel+chanadd, sizeof(float), N+2, stdout );
	fflush( stdout );
	continue;
      }

      leanunconvert( channel+chanadd, buffer, N2 );
      rfft( buffer, N2, INVERSE );
      overlapadd( buffer, N, Wsyn, output, Nw, on );
      shiftout( output, Nw, I, on );
      ++chanindex;
      chanadd += chanoffset;
    }
    exit(0);
}

usage(woof)
{
    fprintf(stderr, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
	"2d:  short-time fourier analysis/resynthesis\n",
	"2d   [flags] < floatsams > floatsams\n",
	"	N:	fft length [2^n]\n",
	"	R:	sampling rate\n",
	"	M:	window size in samples\n",
	"	D:	decimation factor in samples\n",
	"	I:	interpolation factor in samples\n",
	"	P:	oscillator bank pitch factor\n",
	"	f:	time-varying function file\n",
	"	l:	length of input\n",
	"	a:	analysis data output\n",
	"	s:	synthesize analysis input\n",
	"	p:	linear prediction order\n",
	"	g:	oscillator resynthesis gate threshold\n",
	"	h:	this happening place\n");
    exit(woof);
}
