#include <stdio.h>
#include <stdlib.h>
#include "sfheader.h"
#include "crack.h"
#include "pv.h"

#define Max_Table_Size 1048576

typedef struct {
  int           elms;
  float         *data;
} Tabl;

complex zero = {0., 0.};
complex one = {1., 0.};
float PI;
float TWOPI;
float synt = 0.;

main(argc, argv)
    int argc; char *argv[];
{
    int 	i,
		R = 44100,
		N = 1024,
		N2,
		Nw = 0,
		Nw2, 
		D = 0, 
		I = 0,
		in,
		on,
    		freak, ample,
		eof = 0,
		aflag = 0;
    float 	adiff,
		pitch = 1.,
		*Hwin,
		*Wanal,
		*Wsyn,
		*input1,
                *input2,
		*buffer1,
                *buffer2,
		*channel1,
                *channel2,
		*output;
    double	mult = 1.;
    char	ch,
		file1[1024],
                file2[1024],
		mapfile[1024],
		*dbuf;
    FILE	*fp1,
                *fp2,
		*fp3;
    Tabl	map;

    if (argc < 3)
	usage(1);

    while( (ch= crack( argc, argv, "R|N|M|D|I|P|f|F|m|iah", 0  )) != NULL ) {
	switch(ch) {
	    case 'R':	R = atoi(arg_option);
			break;
	    case 'N':	N = atoi(arg_option);
			break;
	    case 'M':	Nw = atoi(arg_option);
			break;
	    case 'D':	D = atoi(arg_option);
			break;
	    case 'I':	I = atoi(arg_option);
			break;
	    case 'P':	pitch = atof(arg_option);
			break;
	    case 'f':	strcpy(file1, arg_option);
			break;
	    case 'F':	strcpy(file2, arg_option);
			break;
	    case 'm':	strcpy(mapfile, arg_option);
			break;
	    case 'a':	aflag = 1;
			break;
	    case 'h':	usage(1);
	}
    }

    if (Nw == 0)
	Nw = N;

    if (D == 0)
      D = N / 8;

    if (aflag)
	I = 0;
    else
	if (I == 0)
	    I = D;

    PI = 4.*atan(1.);
    TWOPI = 8.*atan(1.);
    N2 = N>>1;
    Nw2 = Nw>>1;

    Wanal = (float *) space( Nw, sizeof(float) );
    Wsyn = (float *) space( Nw, sizeof(float) );
    Hwin = (float *) space( Nw, sizeof(float) );
    input1 = (float *) space( Nw, sizeof(float) );
    input2 = (float *) space( Nw, sizeof(float) );
    output = (float *) space( Nw, sizeof(float) );

    buffer1 = (float *) space( N, sizeof(float) );
    buffer2 = (float *) space( N, sizeof(float) );
    channel1 = (float *) space( N+2, sizeof(float) );
    channel2 = (float *) space( N+2, sizeof(float) );

    makewindows( Hwin, Wanal, Wsyn, Nw, N, I, 0 );

    in = -Nw;
    if ( D )
	on = (in*I)/D;
    else
	on = in;

/* open input files */
    if ( (fp1 = fopen( file1, "r" )) == NULL ) {
	fprintf(stderr,"Where is %s?\n",file1);
	exit(-1);
    }
    if ( (fp2 = fopen( file2, "r" )) == NULL ) {
      fprintf(stderr,"Where is %s?\n",file2);
      exit(-1);
    }

    (void) stripheader(fp1);
    (void) stripheader(fp2);

    map.elms = Max_Table_Size;
    map.data = (float *) space( Max_Table_Size, sizeof(float) );
    
    if ( (fp3 = fopen( mapfile, "r" )) == NULL ) {
      fprintf(stderr,"What is %s?\n", mapfile);
      exit(-1);
    }
    i = 0;
    while ( (fread(map.data+i, sizeof(float), 1, fp3)) != 0 && 
	   i < Max_Table_Size )
      ++i;
    map.elms = i;
    
    while ( !eof ) {

	in += D;
	on += I;

	eof = turnin( input1, Nw, D, fp1 );
	if (eof)
	  (void) turnin( input2, Nw, D, fp2 );
	else
	  eof = turnin( input2, Nw, D, fp2 );
	
	fold( input1, Wanal, Nw, buffer1, N, in );
	fold( input2, Wanal, Nw, buffer2, N, in );
	rfft( buffer1, N2, FORWARD );
	rfft( buffer2, N2, FORWARD );
	convert( buffer1, channel1, N2, D, R );
	convert( buffer2, channel2, N2, D, R );
	
	//	  baiter( buffer1, buffer2, channel1, channel2, N2, mult );

	for ( i=0; i < N2; i++ ) {
	  freak = (ample = i << 1) + 1;
	  *(channel1+freak) = abs( (*(channel2+freak) + 
				    *(channel1+freak)) / 2.0);
//	  adiff = *(channel2+ample) / *(channel1+ample);
	  *(channel1+ample) = (*(channel2+ample) + *(channel1+ample)) / 2.0;
	}


	if ( I == 0 ) {
	    fwrite( channel1, sizeof(float), N+2, stdout );
	    fflush( stdout );
	    continue;
	}

        noscbank( channel1, N2, R, Nw, I, pitch, output );
	shiftout( output, Nw, I, on );
    }
    exit(0);
}

void stripheader(fp)
FILE *fp;
{
  register int i;

  for ( i=0; i < SIZEOF_HEADER; i++ )
    fgetc(fp);
}

usage(woof)
{
    fprintf(stderr, "%s%s%s%s%s%s%s%s%s%s%s%s%s",
	"bait:  more spectral interpolation\n",
	"bait   [flags] > floatsams\n",
	"	N:	fft length [2^n]\n",
	"	R:	sampling rate\n",
	"	M:	window size in samples\n",
	"	D:	decimation factor in samples\n",
	"	I:	interpolation factor in samples\n",
	"	P:	frequency multiplier\n",
	"	f:	source soundfile\n",
	"	F:	filter soundfile\n",
	"	m:	amplitude interpolation mapping function\n",
	"	a:	analysis data output\n",
	"	h:	this cozy place\n");
    exit(woof);
}
