#include <stdio.h>
#include "pv.h"
#include "crack.h"

#define Max_Pairs 256 
#define allc_min 4096

typedef struct {
    float min;
    float max;
} Bound;

typedef struct {
    int 	its;
    float	*ival;
    float 	*prob;
} ISstr;

typedef struct {
    float	*tlc;
    int 	tlen;
} Tabl;

complex zero = { 0., 0. } ;
complex one = { 1., 0. } ;
float PI ;
float TWOPI ;
float synt = 0.;

main(argc, argv)
    int argc ; char *argv[] ;
{
    int 	R,
		N,
		N2,
		Nw = 0,
		Nw2, 
		D = 0, 
		I = 0,
		i,j,k,
		hoop,			/* looping variable */
		in,
		on,
		eof = 0,
		obank = 0,
		Np = 0;
    float 	P = 0.,
		xp = 1.,
		ampl = 1.,
		*Hwin,
		*Wanal,
		*Wsyn,
		*input,
		*winput,
		*lpcoef,
		*buffer,
		*channel,
		*output,
		lpa(),
		maxamp();
    char	ch,
		file[128],
		*dbuf;
    FILE	*fp;
    ISstr	*is;

    if (isatty(0))
	usage(1);

    while( (ch= crack( argc, argv, "R|N|M|D|I|P|e|a|p|g|h", 0  )) != NULL ) {
	switch(ch) {
	    case 'R':	R = atoi(arg_option);
			break;
	    case 'N':	N = atoi(arg_option);
			break;
	    case 'M':	Nw = atoi(arg_option);
			break;
	    case 'D':	D = atoi(arg_option);
			break;
	    case 'I':	I = atoi(arg_option);
			break;
	    case 'P':	P = atof(arg_option);
			break;
	    case 'e':	xp = atof(arg_option);
			break;
	    case 'a':	ampl = atof(arg_option);
			break;
	    case 'p':	Np = atoi(arg_option);
			break;
	    case 'g':	synt = atof(arg_option);
			break;
	    case 'h':	usage(1);
	}
    }

    if (Nw == 0)
	Nw = N;

    PI = 4.*atan(1.) ;
    TWOPI = 8.*atan(1.) ;
    obank = P != 0. ;
    N2 = N>>1 ;
    Nw2 = Nw>>1 ;
/*
 * allocate memory
 */
    fvec( Wanal, Nw ) ;		/* analysis window */
    fvec( Wsyn, Nw ) ;		/* synthesis window */
    fvec( input, Nw ) ;		/* input buffer */
    fvec( Hwin, Nw ) ;		/* plain Hamming window */
    fvec( winput, Nw ) ;	/* windowed input buffer */
    fvec( lpcoef, Np+1 ) ;	/* lp coefficients */
    fvec( buffer, N ) ;		/* FFT buffer */
    fvec( channel, N+2 ) ;	/* analysis channels */
    fvec( output, Nw ) ;	/* output buffer */
/*
 * create windows
 */
    makewindows( Hwin, Wanal, Wsyn, Nw, N, I, obank ) ;
/*
 * initialize input and output time values (in samples)
 */
    in = -Nw ;
    if ( D )
	on = (in*I)/D ;
    else
	on = in;

    is = (ISstr *) space( 1, sizeof(ISstr*) );
    is->ival = (float *) space( Max_Pairs, sizeof(float) );
    is->prob = (float *) space( Max_Pairs, sizeof(float) );

    if ( (fp = fopen( file, "r" )) == NULL ) {
	fprintf(stderr,"what is %s?", file);
	exit(-1);
    }

    j=0;
    while ( fscanf(fp, "%g %g\n", *(is->ival+j), *(is->prob+j) != EOF ) 
	++j;
    if ( j==0 ) {
	fprintf(stderr,"No interval data!\n");
	exit(-1);
    }
    is->its = j;

/*
 * main loop--perform phase vocoder analysis-resynthesis
 */
    while ( !eof ) {
/*
 * increment times
 */
	in += D ;
	on += I ;
/*
 * analysis: input D samples; window, fold and rotate input
 * samples into FFT buffer; take FFT; and convert to
 * amplitude-frequency (phase vocoder) form
 */
	if ( D == 0 ) {
	    if (Np) {
		if ( fread( lpcoef, sizeof(float), Np+1, stdin ) == 0 )
		    eof = 1;
	    }
		if ( fread( channel, sizeof(float), N+2, stdin ) == 0 )
		    eof = 1;
	}		
	else {
	    eof = shiftin( input, Nw, D ) ;
	    if ( Np ) {
		vvmult( winput, Hwin, input, Nw ) ;
		lpcoef[0] = lpa( winput, Nw, lpcoef, Np, 0 ) ;
	    }
	    fold( input, Wanal, Nw, buffer, N, in ) ;
	    rfft( buffer, N2, FORWARD ) ;
	    convert( buffer, channel, N2, D, R ) ;
	}
	if ( I == 0 ) {
	    if ( Np )
		fwrite( lpcoef, sizeof(float), Np+1, stdout ) ;
	    fwrite( channel, sizeof(float), N+2, stdout ) ;
	    fflush( stdout ) ;
	    continue ;
	}
	marx(channel, is, N);
/*
 * at this point channel[2*i] contains amplitude data and
 * channel[2*i+1] contains frequency data (in Hz) for phase
 * vocoder channels i = 0, 1, ... N/2; the center frequency
 * associated with each channel is i*f, where f is the
 * fundamental frequency of analysis R/N; any desired spectral
 * modifications can be made at this point: pitch modifications
 * are generally well suited to oscillator bank resynthesis,
 * while time modifications are generally well (and more
 * efficiently) suited to overlap-add resynthesis
 */
	if ( obank ) {
/*
 * oscillator bank resynthesis
 */
	    oscbank( channel, N2, lpcoef, Np, R, Nw, I, P, output ) ;
	    shiftout( output, Nw, I, on+Nw-I ) ;
	} else {
/*
 * overlap-add resynthesis
 */
	    unconvert( channel, buffer, N2, I, R ) ;
	    rfft( buffer, N2, INVERSE ) ;
	    overlapadd( buffer, N, Wsyn, output, Nw, on ) ;
	    shiftout( output, Nw, I, on ) ;
	}
    }
    exit( 0 ) ;
}
vvmult( out, a, b, n ) float *out, *a, *b ; int n ; {
 register float *lim = out + n ;
    while ( out < lim )
	*out++ = *a++ * *b++ ;
}
float maxamp( in, n, per, ix, sz ) float *in ; int n, per, ix, sz ; {
 char *malloc(), *calloc() ;
 float maxof(), *lim = in + n ;
 int i, found ;
 static float **buf, *mx ;
 static int *ptr ;
 static int first = 1 ;
 register int *p ;
 register float *b, *m ;
    if ( first ) {
	first = 0 ;
	buf = (float **) malloc( sz*sizeof(float *) ) ;
	for ( i = 0 ; i < sz ; i++ )
	    buf[i] = (float *) calloc( per, sizeof(float) ) ;
	ptr = (int *) calloc( sz, sizeof(int) ) ;
	mx = (float *) calloc( sz, sizeof(float) ) ;
    }
    --ix ;
    for (
	b = buf[ix] , p = &ptr[ix] , m = &mx[ix], found = 0 ;
	in < lim ;
	in++
    ) {
	if ( ( *(b+*p) = *in >= 0. ? *in : -*in ) >= *m )
	    found = *m = *(b+*p) ;
	if ( ++(*p) > per )
	    *p = 0 ;
    }
    if ( found )
	return( *m ) ;
    else
	return( *m = maxof( b, per ) ) ;
}
float maxof( a, n ) float *a ; int n ; {
 register float *lim = a + n, m ;
    for ( m = *a++ ; a < lim ; a++ )
	if ( *a > m )
	    m = *a ;
    return( m ) ;
}

marx( cur, is, n, look )
float *cur; ISstr *is; int n; Tabl look;
{
    register int i,j;
    float prand();
    float rn;
    static float *sft = NULL;

    if (*sft == NULL) {
	fvec(sft, n);
	for( i=3; i < n; i+=2 ) {
	    rn = prand();
	    j = 0;
	    while ( rn -= *(is->prob+j) > 0. )
	    	++j;
	    if ( prand() < .5 )
		*(sft+i) = *(is->prob+j);
	    else
		*(sft+i) = 1. / *(is->prob+j);

    for( i=3; i < n; i+=2 ) {
    }	

	
	
}

Tabl *makelook( is )
STstr *is; int its;
{
    register int i, j = 0, index;
    static Tabl *look;

    for( i=0; i < is->its; i++ ) {
	index = *(is->prob+i) * ;
	for( j=0; j <  

}


Bound extrema( n, data )
int n; float *data;
{
    register int i;
    Bound ext;

    ext.min = *data;
    for( i=2; i < n; i+=2 ) {
	if ( ext.min > fabs(*(data+i)) )
	    ext.min = fabs(*(data+i));
    }
    ext.max = *data;
    for( i=2; i < n; i+=2 ) {
	if ( ext.max < fabs(*(data+i)) )
	    ext.max = fabs(*(data+i));
    }
    return ext;
}

char *space( sz, obsz )
int sz, obsz;
{
    char	*p;

    if ( (sz * obsz) < allc_min ) {
	if ( (p = (char *) malloc( allc_min )) == NULL ) {
	    fprintf(stderr,"No Memory!\n");
	    exit(-1);
	}
    }
    else {
	if ( (p = (char *) malloc( sz * obsz )) == NULL ) {
	    fprintf(stderr,"No Memory!\n");
	    exit(-1);
	}
    }
    return p;
}

static long randx = 1;

srrand(x)
unsigned x;
{
    randx = x;
}

float rrand()
{
    int i = ((randx = randx*1103515245 + 12345)>>16) & 077777;
    return((float)i/16384. - 1.);
}

float prand()
{
    int i = ((randx = randx*1103515245 + 12345)>>16) & 077777;
    return((float)i/32768.);
}

usage(woof)
{
    fprintf(stderr, "%s%s%s%s%s%s%s%s%s%s%s",
	"marx: spectral redistribution processor\n",
	"marx  [flags] < floatsams > floatsams\n",
	"	N:	fft length [2^n]\n",
	"	R:	sampling rate\n",
	"	M:	window size in samples\n",
	"	D:	decimation factor in samples\n",
	"	I:	interpolation factor in samples\n",
	"	P:	oscillator bank pitch factor\n",
	"	f:	interval data file\n",
	"	p:	linear prediction order\n",
	"	g:	oscillator resynthesis gate threshold\n");
    exit(woof);
}
