#include <stdlib.h>
#include <stdio.h>
#include "pv.h"
#include "crack.h"

complex zero = { 0., 0. };
complex one = { 1., 0. };
float PI;
float TWOPI;
float synt = 0.;

void lexp( cur, xp, ampl, n )
float *cur, xp, ampl; int n;
{
  register int 	i;
  float		dif,
		iexp,
		dexp;
  Bound		faext,			/* pvoc frame extrema */
		extrema();		/* extrema determining function */
   
    iexp = 1. / xp;
    dexp = xp - iexp;
    faext = extrema( n, cur );
    faext.max = (10. * log10(faext.max));
    faext.min = (10. * log10(faext.min));
    dif = faext.max - faext.min;
    for( i=2; i < n; i+=2 )
      *(cur+i) *= ampl * pow( 10., ( ( (10. * log10( *(cur+i) - faext.min )
					/ dif) * dexp ) + iexp));
}

void pexp( cur, xp, ampl, n )
float *cur, xp, ampl; int n;
{
  register int 	i;
  float		dif,
		iexp,
		dexp;
  Bound		faext,			/* pvoc frame extrema */
		extrema();		/* extrema determining function */
   
    iexp = 1. / xp;
    dexp = xp - iexp;
    faext = extrema( n, cur );
    dif = faext.max - faext.min;
    for( i=2; i < n; i+=2 )
	*(cur+i) *= ampl * ((( (*(cur+i) - faext.min) / dif ) * dexp) + iexp);
}

Bound extrema( n, data )
int n; float *data;
{
  register int 	i;
  Bound 	ext;

    ext.min = *data;
    for( i=2; i < n; i+=2 ) {
	if ( ext.min > fabs(*(data+i)) )
	    ext.min = fabs(*(data+i));
    }
    ext.max = *data;
    for( i=2; i < n; i+=2 ) {
	if ( ext.max < fabs(*(data+i)) )
	    ext.max = fabs(*(data+i));
    }
    return ext;
}

main(argc, argv)
    int argc; char *argv[];
{
    int 	R,
		N = 1024,
		N2,
		Nw = 0,
		Nw2, 
		D = 0, 
		I = 0,
		i,
		hoop,			/* looping variable */
		in,
		on,
		eof = 0,
		obank = 0;
    float 	P = 0.,
		xp = 1.,
		ampl = 1.,
		*Hwin,
		*Wanal,
		*Wsyn,
		*input,
		*buffer,
		*channel,
		*output;
    char	ch,
		*dbuf;

    if (isatty(0))
	usage(1);

    while( (ch= crack( argc, argv, "R|N|M|D|I|P|e|a|p|g|h", 0  )) != NULL ) {
	switch(ch) {
	    case 'R':	R = atoi(arg_option);
			break;
	    case 'N':	N = atoi(arg_option);
			break;
	    case 'M':	Nw = atoi(arg_option);
			break;
	    case 'D':	D = atoi(arg_option);
			break;
	    case 'I':	I = atoi(arg_option);
			break;
	    case 'P':	P = atof(arg_option);
			break;
	    case 'e':	xp = atof(arg_option);
			break;
	    case 'a':	ampl = atof(arg_option);
			break;
	    case 'g':	synt = atof(arg_option);
			break;
	    case 'h':	usage(1);
	}
    }

   
    if (Nw == 0)
	Nw = N;

    PI = 4.*atan(1.);
    TWOPI = 8.*atan(1.);
    obank = P != 0.;
    N2 = N>>1;
    Nw2 = Nw>>1;

/* allocate memory */

    Wanal = (float *) space( Nw, sizeof(float) );	/* analysis window */
    Wsyn = (float *) space( Nw, sizeof(float) );	/* synthesis windows */
    input = (float *) space( Nw, sizeof(float) );	/* input buffers */
    Hwin = (float *) space( Nw, sizeof(float) );	/* plain Hamming window */
    buffer = (float *) space( N, sizeof(float) );	/* FFT buffer */
    channel = (float *) space( N+2, sizeof(float) );
    output = (float *) space( Nw, sizeof(float) );	/* output buffer */

/* create windows */

    makewindows( Hwin, Wanal, Wsyn, Nw, N, I, obank );
/* initialize input and output time values (in samples) */

    in = -Nw;
    if ( D )
	on = (in*I)/D;
    else
	on = in;
/* main loop--perform phase vocoder analysis-resynthesis */

    while ( !eof ) {

/* increment times */

	in += D;
	on += I;

/* analysis: input D samples; window, fold and rotate input
   samples into FFT buffer; take FFT; and convert to
   amplitude-frequency (phase vocoder) form */

	if ( D == 0 ) {
	  if ( fread( channel, sizeof(float), N+2, stdin ) == 0 )
	    eof = 1;
	}		
	else {
	    eof = shiftin( input, Nw, D );
	    fold( input, Wanal, Nw, buffer, N, in );
	    rfft( buffer, N2, FORWARD );
	    convert( buffer, channel, N2, D, R );
	}
	if ( I == 0 ) {
	    fwrite( channel, sizeof(float), N+2, stdout );
	    fflush( stdout );
	    continue;
	}
	pexp(channel, xp, ampl, N);

/* at this point channel[2*i] contains amplitude data and
   channel[2*i+1] contains frequency data (in Hz) for phase
   vocoder channels i = 0, 1, ... N/2; the center frequency
   associated with each channel is i*f, where f is the
   fundamental frequency of analysis R/N; any desired spectral
   modifications can be made at this point: pitch modifications
   are generally well suited to oscillator bank resynthesis,
   while time modifications are generally well (and more
   efficiently) suited to overlap-add resynthesis */

/* oscillator bank resynthesis */

	if ( obank ) {
	    noscbank( channel, N2, R, Nw, I, P, output );
	    shiftout( output, Nw, I, on+Nw-I );
	}

/* overlap-add resynthesis */

	else {
	    unconvert( channel, buffer, N2, I, R );
	    rfft( buffer, N2, INVERSE );
	    overlapadd( buffer, N, Wsyn, output, Nw, on );
	    shiftout( output, Nw, I, on );
	}
    }
    exit(0);
}


usage(woof)
{
    fprintf(stderr, "%s%s%s%s%s%s%s%s%s%s%s%s",
	"pexp:  phase vocoder based spectral expansion\n",
	"pexp   [flags] < floatsams > floatsams\n",
	"	N:	fft length [2^n]\n",
	"	R:	sampling rate\n",
	"	M:	window size in samples\n",
	"	D:	decimation factor in samples\n",
	"	I:	interpolation factor in samples\n",
	"	P:	oscillator bank pitch factor\n",
	"	e:	spectral expansion factor\n",
	"	a:	amplitude multiplier\n",
	"	p:	linear prediction order\n",
	"	g:	oscillator resynthesis gate threshold\n",
	"	h:	this gleeful place\n");
    exit(woof);
}
