//#include "macros.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "../H/ugens.h"
#include "../H/sfheader.h"
//#include "setell.h"

#define SIZE 1024

extern SFHEADER sfdesc[2];
extern struct stat sfst[2];
extern int headersize[2];
extern SFMAXAMP sfm[2];
int RESET = 200;  /* times per sec to reinitialize envel */

// prototypes
double ell(float *,int,double *);
float ellipse(double x, double C[4][10],double PS[4][10]);
void ellpset(double *list, double C[4][10], double PS[4][10]);
double ellset(float *p,int n_args,double *);
void profile(void);
double reset(float *,int,double *);
void sl(float *,int,double *);

void setell(double, double, double, double, double, double,double *, int *,short);
extern double setline(float *,int ,double *);
extern double panline(float *,int ,double *);
extern void tableset(float,int,float *);
extern float tablei(int ,float *,float *);
extern int setnote(double,double,int);
extern double m_dur(float *,int, double *);
extern endnote(int);

//------------------------------
float  envel, pan;
double psarr[4][4][10],carr[4][4][10],xnorm;
double retarr[461];
float parray[SIZE];
float sarray[SIZE];
float tabs[2]; /* for setline */
float ptabs[2]; /* for pan */
float *array;


/* global array for coeffeicients */
int	nsects, /* number of sections */
		count,
		input = 0;
int lineset = 0;
int panflag = 0;

double
ellset(float *p,int n_args,double *pp){
	double sampr,f1,f2,f3,ripple,atten;

/* pp0=passband cutoff;
 pp1=stopband cutoff;
 pp2=0 or bandpass stopband;
 pp3=ripple;
 pp4=attenuation(db);
 pp5=verbose
*/
	p=p;n_args=n_args;
	f1 = pp[0];
	f2 = pp[1];
	f3 = pp[2];
	sampr = SR;
	ripple = pp[3];
	atten = pp[4];
	setell(sampr,f1,f2,f3,ripple,atten,retarr,&nsects,(short)pp[5]);
	return(0);
}

double
ell(float *p,int n_args,double *pp){
/*
	pp0 = inskip,
	pp1=outskip,
	pp2=duration. 0 is total dur, -[val] is dur, [val] is endtime 
	pp3=chanin,
	pp4=chanout,
	pp5=amp
 */	

int i,j,nsamps,nchans;
float x[4],y[4], amp, tmpx;
double dur;

	p=p;
//fprintf(stderr,"dur %f !\n",pp[2]);
	if(!pp[2]) pp[2] = m_dur(&p[2],1,&pp[2]);
	dur = (pp[2] < 0) ? -pp[2] : (pp[2] - pp[0]);
//fprintf(stderr,"dur %f !\n",dur);
	nsamps=setnote(pp[0], dur,0);
  setnote(pp[1],dur,1);
	nchans = sfchans(&sfdesc[input]);

	if ( sfchans(&sfdesc[1]) < 2 && panflag ) {
		fprintf(stderr,"output must be 2 channels for pan active!\n");
		exit(-1);
		}
	if (n_args<6) amp=1.0;
	else amp=(float)pp[5];

	if(!lineset) {
		for(j=0; j<SIZE; j++) parray[j] = sarray[j] = 1.0;
			fprintf(stderr,"Set phrase curve to all 1's\n");
			lineset = 1;
	}

	tableset((float)dur,SIZE,tabs);
	if (panflag){ tableset((float)dur,SIZE,ptabs);}
	count = SR/RESET;

	for(j=0;j<nchans;j++) ellpset(retarr, carr[j], psarr[j]);
	for(i=0;i<4;i++) x[i] = y[i] = 0.0;

	if(!panflag) {
		for(i=0;i<nsamps;i++) {
			while(!j--) {
				j=count;
				envel = tablei(i,sarray,tabs);
			}
			GETIN(y,0);
			j=0;
			x[(int)pp[4]] = ellipse((double)y[(int)pp[3]], carr[j], psarr[j]) * envel * amp;
 			ADDOUT(x,1);
 			}
		}
	else {
		for(i=0;i<nsamps;i++) {
			while(!j--) {
				j=count;
				pan = tablei(i,parray,ptabs);
				envel = tablei(i,sarray,tabs);
			}
			GETIN(y,0);
			j=0;

//printf("envel %f. pan %f\n",envel, pan);
			tmpx = ellipse((double)y[(int)pp[3]], carr[j], psarr[j]) * envel * amp;
			x[0] = tmpx * (1.0 - pan);
			x[1] = tmpx * pan;

//			for(j=0;j<nchans;j++) {
//				x[j] = ellipse(y[j], carr[j], psarr[j]) * amp * envel;}

 			ADDOUT(x,1);
		}
	}
 	endnote(1);
	return(0);
}

float ellipse(double x, double C[4][10],double PS[4][10]){
	int m;
	double op;
	for(m=0;m<nsects;m++) {
		op = x + C[0][m] * PS[0][m] + C[2][m] * PS[1][m]-C[1][m] * PS[2][m] - C[3][m] * PS[3][m];
		PS[1][m] = PS[0][m];
		PS[0][m] = x;
		PS[3][m] = PS[2][m];
		PS[2][m] = op;
		x = op;
		}
	return((float)(x*xnorm));
}

void ellpset(double *list, double C[4][10], double PS[4][10]){
/* the first argument in the list is the number of sections */
/* this is no longer true -- nsects contains number of sections -- BGG */
int m,i,j;
i=0;
	for(m=0;m<nsects;m++) {
		for(j=0;j<4;j++)  {
    	C[j][m] = list[i++];
			 PS[j][m] = 0;
	}		
 }	xnorm = list[i];
	printf("number of sections in this filter: %d\n",nsects);
	printf("normalization factor: %f 	number of terms: %d\n",xnorm,i);
}

void sl(float *p,int n_args,double *pp)
{
	p=p;
//	setline(pp,n_args,SIZE,array);
	array=sarray;
	setline(p,n_args,pp);
	lineset=1;
}

double
panline(float *p,int n_args, double *pp)
{
	double increm;
	int i,j,k,points;

	p=p;
	if((n_args % 2) != 0) {
		fprintf(stderr,"Something wrong with phrase, check args for mix setline\n");
		closesf();
	}

	lineset = 1;   /* initialized array */
	panflag=1;
	increm = (pp[n_args - 2] - pp[0])/(double)SIZE;
	for(j=0,i=0; j < (n_args-2); j += 2) {
		points = (int)((pp[j+2] - pp[j]) / increm +.5);
		if(pp[j+2] != pp[j]) {
			if(points <= 0) points = 1;
			if((pp[j+2] < pp[j]) || (points > SIZE)) {
				fprintf(stderr," confusion on phrase card\n");
				closesf();
				}
			for(k=0; k < points; k++) {
				parray[i++] = ((float)k/(float)points)
					* (pp[j+3] - pp[j+1]) + pp[j+1];
				if(i == SIZE) return(0.0);
			}
		}
	}
	i--;
	while(++i < SIZE) parray[i] = parray[i-1]; 
	return(0.0);
}

double
reset(float *p,int n_args,double *pp)
{
	p=p; n_args=n_args;
	if(pp[0]) RESET = pp[0];
	fprintf(stderr,"envel calls set to %d times per sec\n",RESET);
  return 0;
}
