#include "f2c.h"
#include <stdio.h>

/*
COMPLEX FUNCTIONS FOR USE WITH ELLIPTICAL
FILTER DESIGN PROGRAM. CODED BY ERIC LYON. 
*/


/* place z2 ** 2 into z1 */
int cdsq_(z1,z2)
doublecomplex *z1, *z2;
{
	z1->r = z2->r * z2->r - z2->i * z2->i;
	z1->i = 2. * z2->r * z2->i;
}

/* place complex sqrt(z2) into z1 */
int cdsqrt_(z1,z2)
doublecomplex *z1, *z2;
{
double r, angle;
double sin(), cos(), atan(),atan2();
doublereal cdabs_();
double sqrt();
double pi;
double epsilon = .1E-30;
double fabs();
double pio2;
pi = 4. * atan(1.);
pio2 = 2. * atan(1.);

r = cdabs_(z2);
if( fabs(z2->r) < epsilon )
	angle =  pio2 ;
else 
	angle = atan2( z2->i,  z2->r );
z1->r = sqrt(r) * cos( angle/2.0 );
z1->i = sqrt(r) * sin( angle/2.0 );

}

/* return real part of complex number */

double dreal_(x)
doublecomplex *x;
{
	return( x->r );
}

/* place exp(z2) into z1 */
#define E_ITERS (20)
int cdexp_(z1,z2)
doublecomplex *z1, *z2;
{
double radius, angle, atan(), pow();
double sin(), cos();
doublereal cdabs_();
int factorial();
doublecomplex ztmp;
double faci ;
int i, j ;
doublereal r_to_n ;
	angle = atan( z2->i / z2->r );
	radius = cdabs_(z2);
	z1->i = z2->i;
	z1->r = 1.0 + z2->r;
	if( z2->i == 0.0 && z2->r == 0.0 ){
		z1->i = 0.0;
		z1->r = 1.0;
	}
	else
	for( i = 2; i < E_ITERS; i++ ){
		r_to_n = pow(radius, (double) i);				ztmp.r = r_to_n * cos( (double)i * angle);
		ztmp.i = r_to_n * sin( (double)i * angle); 
		faci = (double) factorial(i);
		if( faci == 0. ){
			break;
		}
		ztmp.r /= faci;
		ztmp.i /= faci;
		z1->i += ztmp.i;
		z1->r += ztmp.r; 
	}
}

/* return factorial of x */

factorial( x ) int x;
{
int fac = x;
	while( x-- )
		if( x > 0 )
			fac *= x;
	return( fac );
}

/* return modulus or abs value of complex number */

doublereal cdabs_(x)
doublecomplex *x;
{	
	double sqrt();
	return( sqrt( x->r * x->r + x->i * x->i ) );
}
