/*
 * modified - Fri May  3 21:11:42 EST 1996
 *
 * c incarnation of random-wait chorus instrument 
 * p0 = start,p1=dur,p2=inskip,p3=indur,p4=nvoices, p5=amlow, p6=amphi
 * p7&8 = range of randomwait, p9&10 = rise and decay times for segment
 * p11 = seed
 *
 * rise/decay for each voice can be implemented by calling slv()
 * sl() is a global setline
*/

#include "sfheader.h"
#include "ugens.h"
#include "macros.h"
#include <stdio.h>
#include <unistd.h>

#define MAXVOICES 100
#define SIZE 1024
//#define MAXSIZE 8192
#define MAXSIZE 32767
#define ARRAYSIZE 1024

extern float *floc(int genno);
extern void evset(float dur,float rise,float dec,int nfrise,float *q);
extern float evp(int nsample,float *f1, float *f2, float *q);
extern int setnote(double,double,int);
extern tableset(float,int,float *);
extern double crandom(double);
extern double setline(float *,int, double *);

void clockset(float *p,int n_args,double *pp);
void sl(float *p,int n_args,double *pp); // global setline
void slv(float *p,int n_args,double *pp); // individual voice setline

float array[ARRAYSIZE];
float amp_array[ARRAYSIZE];
float amp2_array[ARRAYSIZE];
//float *array;
int clock_reset=200;
int slv_flag=0;
int lineset=0;

double
chorus(float *p,int n_args,double *pp)
{
float ampout;
float ampvals[2];
float *signal;
float evals[5];
float xin[2],out[2];  /* designed for only first channel though */
float *f1;
float sig;
double seed;
int i,j,input,output,nsamps,nvoices,wait[MAXVOICES],limit;
float sloc[MAXVOICES],srite[MAXVOICES],amp[MAXVOICES];
float ampdiff,ampbot,waitdiff,waitbot;
int count,clock;

f1 = floc(0); /* rise function is in floc 0 */
p=p;n_args=n_args;
input=0; 
output=1;
evset((float)pp[3],(float)pp[9],(float)pp[10],0,evals);
ampdiff = (float)pp[6] - (float)pp[5];
ampbot = (float)pp[5];
waitdiff = (float)((float)pp[8] - (float)pp[7]) * SR;
waitbot = (float)pp[7] * SR;
//nsamps = setnote(pp[0],pp[1],output);
nsamps = setnote(pp[2],pp[3],input);
tableset((float)pp[1],ARRAYSIZE,ampvals);
count = 0;
clock = clock_reset;

if((signal = (float *)calloc(nsamps,sizeof(float))) == 0) {
//if((signal = (float *)malloc(nsamps*4)) == 0) {
  printf("can't get enough memory \n");
  closesf();
}

for(i=0; i<nsamps; i++)  {
	GETIN(xin,input);
	if(slv_flag){
		if(!count--) {
			ampout = table(i,amp2_array,ampvals);
			count = clock;
		}
		signal[i] =  xin[0] * ampout;
	}
	else {
	signal[i] =  xin[0];
//	signal[i] =  xin[0] * evp(i,f1,f1,evals);
	}
}
nvoices = pp[4];
seed=pp[10];
limit = nsamps - 1;
seed = (float)pp[11];
for(i=0; i<nvoices; i++) {
  wait[i] = (float)(-i/(float)(nvoices-1)) * limit;
  seed = crandom(seed);
	sloc[i] = (float)seed;
	srite[i] = 1. - sloc[i];
	seed = crandom(seed);
	amp[i] = (float)seed * ampdiff * ampbot;
}
nsamps = setnote(pp[0],pp[1],output);
limit = nsamps - 1; // readjust limit?

count = 0;
for(i=0; i<nsamps; i++,out[0]=out[1]=0.0) {
	for(j=0; j<nvoices; j++) {
		if(!count--) {
			ampout = tablei(i,amp_array,ampvals);
//			printf("ampout %f\n",ampout);
			count = clock;
		}
		if(wait[j]++ < 0) continue;
		if(wait[j] > limit) {
			seed = crandom(seed);
			wait[j] = -seed * waitdiff + waitbot;
			seed = crandom(seed);
			sloc[j] = (float)seed;
			srite[j] = 1. - sloc[j];
			seed = crandom(seed);
			amp[j] = (float)seed * ampdiff + ampbot;
		}
		else{
			sig = signal[wait[j]] * amp[j];
//			out[0] += sig * sloc[j] * ampout;
//			out[1] += sig * srite[j] * ampout;
//			out[0] += sig * sloc[j];
//			out[1] += sig * srite[j];
			out[0] += sig ;
			out[1] += sig;
		}
	}
	ADDOUT(out,output);
}
endnote(output);
free(signal);
return 0;
}	

void
clockset(float *p,int n_args,double *pp)
{
	p=p;n_args=n_args;
	clock_reset = (int)pp[0];
//	fprintf(stderr,"clocking set to %d\n",clock_reset);
}

void
sl(float *p,int n_args,double *pp) // global setline
{
	p=p;
	setline(p,n_args,pp);
// setline(p,n_args,pp,amp_array);

}

void
slv(float *p,int n_args,double *pp) // individual voice setline
{
	p=p;
	slv_flag=1;
//	setline(pp,n_args,ARRAYSIZE,amp2_array);
	setline(p,n_args,pp);
}
