#include "../H/ugens.h"
#include "../H/sfheader.h"
#include <stdio.h>
#include <sys/file.h>
#include <sys/types.h>

extern SFHEADER sfdesc[NFILES];

float ps[4][10],c[4][10],xnorm;
float retarr[461];	/* global array for coefficients */
float sampr,f1beg,f2beg,f3beg,ripplebeg,attenbeg;
float f1end,f2end,f3end,rippleend,attenend;
int nsects;		/* number of sections */

begellset(float p, int n_args,double *pp)
{

/* p0=passband cutoff; p1=stopband cutoff; p2=0 or bandpass stopband;
   p3=ripple; p4=attenuation(db) 	*/

	f1beg = pp[0];
	f2beg = pp[1];
	f3beg = pp[2];
	sampr = SR;
	ripplebeg = pp[3];
	attenbeg = pp[4];
	return 0.0;
}

endellset(float *p,int n_args, double *pp)
{

	f1end = pp[0];
	f2end = pp[1];
	f3end = pp[2];
	sampr = SR;
	rippleend = pp[3];
	attenend = pp[4];
	return 0.0;
}

ell(float *p,int n_args, double *pp)
{

/* p0=inputfile skip, p1=output file skip, p2=duration, p3=frame size (time), p4=beveltime, p5=number of input channel, p6=number of output channel */

	int   r,size,bevsamps,shift;
	long  i,j,frames,framesize,sampno;
        int   k,nchin,nchout,outframesize;
	float *inarr,*outarr,*bevfunc,q[9],z,val;
//	char  *malloc();
        float f1inc,f2inc,f3inc,ripinc,atinc;
	float ellipse();
        float f1,f2,f3,ripple,atten;
	int   inchans, outchans;
	int   amploop,out_file;

	setnote(pp[0],pp[2],0);     /* set input and output file pointers */
        setnote(pp[1],pp[2],out_file);

	inchans = sfchans(&sfdesc[0]);
	outchans = sfchans(&sfdesc[1]);

        bevsamps = pp[4]*SR;       /* samples in beveltime */
	framesize = (pp[3]*SR+(bevsamps*2))*inchans;
        frames = pp[2]/(pp[3]+pp[4]); /*number of filter windows*/

	nchin = pp[5];             /* input and output channels */
	nchout = pp[6];
	f1inc=(f1end-f1beg)/(frames-1);              /* increments */
	f2inc=(f2end-f2beg)/(frames-1); 
	f3inc=(f3end-f3beg)/(frames-1);
	ripinc=(rippleend-ripplebeg)/(frames-1); 
	atinc=(attenend-attenbeg)/(frames-1);

	f1=f1beg;
	f2=f2beg;
	f3=f3beg;
	ripple=ripplebeg;
	atten=attenbeg;
	out_file = 1;

	if ( (inarr = (float *)malloc(framesize*FLOAT)) == NULL ) { 
	          fprintf(stderr,"error in allocating input array. \n");
	          exit(-2);
             }

	outframesize = ((float)framesize) * ((float)outchans/(float)inchans);
	if ( (outarr = (float *)malloc(outframesize*FLOAT)) == NULL ) {
	          fprintf(stderr,"error in allocating output array. \n");
	          exit(-2);
          }

	bevfunc  = floc(1);
	evset(pp[3]+(2*pp[4]),pp[4],pp[4],1,q);    /* set envelope for bevel */
	shift = -bevsamps;            /* no. samples to shift pointers back */

       	for(i=0; i<frames; i++)  {
		nsects = 0;

	 	/* set elliptical filter coefficients */
//		setell_(&sampr,&f1,&f2,&f3,&ripple,&atten,retarr,&nsects);
		setell(sampr,f1,f2,f3,ripple,atten,retarr,&nsects,(float)pp[5]);
		ellpset(retarr);

		BGETIN(inarr,0,framesize);
		amploop = 100;
		for(j=nchin,k=nchout; j<framesize; j+=inchans,k+=outchans) {
			while(!amploop--) {
				val = evp(j/inchans,bevfunc,bevfunc,q);
				amploop = 100;
				}

			outarr[k] = val * ellipse(inarr[j]);
			}

		BADDOUT(outarr,out_file,outframesize);
//		    ADDOUT(outarr,out_file);

		/* reposition input and output pointers */
		inrepos(shift,0);
		outrepos(shift,1);


 //printf("frame %d of %d: %f %f %f %f %f\n",i,frames,f1,f2,f3,ripple,atten);
		/* increment filter stopband & cutoff */
		f1=f1+f1inc;            
		f2=f2+f2inc;
		f3=f3+f3inc;
		ripple=ripple+ripinc;
		atten=atten+atinc;
	}
        endnote(1);
	return 0.0;
}

float ellipse(x)
float x;
{
	register int m;
	float op;

	for(m=0;m<nsects;m++) {
		op = x + c[0][m] * ps[0][m] + c[2][m] * ps[1][m]
		       - c[1][m] * ps[2][m] - c[3][m] * ps[3][m];
		ps[1][m] = ps[0][m];
		ps[0][m] = x;
		ps[3][m] = ps[2][m];
		ps[2][m] = op;
		x = op;
		}
	return(x*xnorm);
}

ellpset(float *list)
{
/* the first argument in the list is the number of sections */
/* this is no longer true -- nsects contains number of sections -- BGG */
	int m,i,j;

	i=0;
	for(m=0;m<nsects;m++) {
		for(j=0;j<4;j++)  {              
	        	 c[j][m] = list[i++];
			 ps[j][m] = 0; }
		}
	xnorm = list[i];
	printf("number of sections in this filter: %d\n",nsects);
	printf("normalization factor: %f  - number of terms: %d\n",xnorm,i);
	return 0.0;
}
