/* cresample.c - cmix sampling rate conversion program
* Cmix version mods - Alistair Riddell
*	DLM - Mon Sep 18 23:16:49 EST 1995
* 
* Notes. When a non-zero is place in pp[9] the program will look for a file
*	called "myfilter" in the current directory and die if it doesn't find it.
*	This was a simple solution to the getting cmix to look for a file without
*	needing to put it in the script. There is probably a more user friendly
* way but at least this is simple.
*/

#include "cresample.h"
#include "../H/ugens.h"
#include "../H/sfheader.h"
#include <stdio.h>
//#include <math.h>
#include "filterkit.h"
//#include <sound/sound.h>
#include "cresample.h"
#include "smallfilter.h"
#include "largefilter.h"
#define IBUFFSIZE 8192		/* Input buffer size */

//--------- Prototypes ---------------------

extern int setnote(float,float,int);	/* position output file*/
void profile(void);
static void fail(char *s);
double resample(float *p,int n_args,double *pp);

static int readData(
				HWORD *outPtr1, // X1 temp buffers
		    HWORD *outPtr2, // X2 temp buffers
		    int dataArraySize, 
		    int nChans, int Xoff);

static int SrcUp(HWORD X[], float Y[], double factor, UWORD *Time,
		 UHWORD Nx, UHWORD Nwing, UHWORD LpScl,
		 HWORD Impp[], HWORD ImpD[], BOOL Interp);

static int SrcUD(HWORD X[], float Y[], double factor, UWORD *Time,
		 UHWORD Nx, UHWORD Nwing, UHWORD LpScl,
		 HWORD Imp[], HWORD ImpD[], BOOL Interp);

int doresample(		/* number of output sample returned */
   double factor,		/* factor = Sndout/Sndin */
   int outCount,		/* number of output samples to compute */
   int nChans,			/* number of sound channels (1 or 2) */
   BOOL interpFilt,		/* TRUE means interpolate filter coeffs */
	int linearInterp,
	BOOL largeFilter,		/* TRUE means use 65-tap FIR filter */
   char *filterFile		/* NULL for internal filter, else filename */);

static int resampleWithFilter(  /* number of output sample returned */
    double factor,		/* factor = Sndout/Sndin */
    int outCount,		/* number of output samples to compute */
    int nChans,			/* number of sound channels (1 or 2) */
    BOOL interpFilt,		/* TRUE means interpolate filter coeffs */
    HWORD Imp[], HWORD ImpD[],
    UHWORD LpScl, UHWORD Nmult, UHWORD Nwing);

static int resampleFast(  /* number of output samples returned */
    double factor,		/* factor = Sndout/Sndin */
    int outCount,		/* number of output samples to compute */
    int nChans);			/* number of sound channels (1 or 2) */

static int err_ret(char *s);

//------------------------------------------------------------------------
extern	SFHEADER	sfdesc[NFILES];  /* to access header information */
int NBYTES = 32768;
char cmix_resampleVERSION[] = "2.1 (Thu Sep 14 22:44:47 EST 1995 - amr)";
int inCount, outCount, outCountReal;
int 	input = 0, output = 1;
float xin[4];
int OBUFFSIZE;
HWORD *X1, *X2;
float  *Y1, *Y2; /* I/O buffers */
float xout[4]; /* std cmix output */
float *Xout; /* std cmix output */
static int trace = 1;		/* for debugging */

#define USAGE "USAGE: resample(p0=inskip, p1=outskip, p2=duration, p3=factor, p4=new rate,"

#define USAGE2 " p5=quality, p6=linearSignalInterp, p7=noInterpolation, p8=filter, p9=terse)"

#define USAGE3 "Note: Either arg 3 or arg 4 must be non-zero."

static void fail(char *s)
{
    fprintf(stderr,"\n\nresample: %s \n\n",s); /* Display error message  */
    fprintf(stderr,USAGE);
	fprintf(stderr,"\n");
    fprintf(stderr,USAGE2);
    fprintf(stderr,"\n\n");
    fprintf(stderr,USAGE3);
    fprintf(stderr,"\n");
    exit(1);			/* Exit, indicating error */
}

double
resample(float *p,int n_args, double *pp)
{
	double factor = -2.0;	/* factor = Sndout/Sndin */
	double newsrate=0;
	BOOL interpFilt = TRUE;	/* TRUE means interpolate filter coeffs */
	BOOL largeFilter = FALSE;	/* TRUE means use 65-tap FIR filter */
	BOOL linearInterp = FALSE;	/* TRUE => no filter, linearly interpolate */
	BOOL knowFactor = FALSE;	/* Used to detect insufficient command-line spec */
	int nChans,outCount;
	char *filterFile; // this should be user friendly
	int i;
	/* Set defaults */

	largeFilter = FALSE;
	factor = -2.0;
	interpFilt = TRUE;

	/*p0=inskip, p1=outskip, p2=duration, p3=factor,p4=new rate, p5=quality, 
	p6=linearSignalInterp, p7=noInterpolation, p8=filter, p9=terse)
	*/

	if (n_args > 10||n_args<10)
		fail("Not the right number of args. Should be 9.");

	trace = pp[9];
	if (trace) printf("Verbose mode is ON\n");
	else printf("Verbose mode is OFF\n");

	if(pp[5]>0.0){			       /* -aaa */
	    largeFilter = TRUE;
	    if (trace) printf("Choosing higher quality filter.\n");
	}
	if(pp[3]>0.0){			       /* -by factor */
	    factor=pp[3];
	    knowFactor = TRUE;
	    if (trace) printf("Sampling-rate conversion factor set to %f.\n",factor);
	}
  if(pp[4]>0.0){            /* -to new srate */
		newsrate=pp[4];
		if (newsrate<sfsrate(&sfdesc[input])/4||newsrate>sfsrate(&sfdesc[input])*4)
			fail("Perhaps, you should use arg 3 instead?");
		if (trace)
			printf("Target sampling-rate set to %f.\n",newsrate);
    if (newsrate>0) {
			if (factor>0)
	 	 		fprintf(stderr,"arg 3 conversion-factor ignored ..."
		 		 " the '-to' option overrides the '-by' option.\n");
				factor = newsrate/sfsrate(&sfdesc[input]);
	  		knowFactor = TRUE;
			if (trace)
	 	   printf("Sampling rate conversion factor set to %f\n",factor);
   	 }
			if (factor < 0) {
				factor = -factor;
				factor = GetDouble("Sampling-rate conversion factor",factor,"");
			}
		}
	if(pp[6]>0){			       /* -linearInpterpolation */
		linearInterp = TRUE;
		if (trace)
		printf("Using linear instead of bandlimited interpolation (faster?)\n");
	}
	if(pp[7]>0){			       /* -noInterpolation */
	  interpFilt = FALSE;
	  if (trace)
	    printf("Filter-table interpolation disabled.\n");
	}
	if(pp[8]!=0){			       /* -filter filterFile */
		i = (int)pp[8];
		filterFile =(char *)i;
	  if (trace) printf("Filter file set to %s.\n",filterFile);
	 }
	else {
	 	filterFile =NULL;
	 	printf("Default Filter data.\n");
	}
  if (!knowFactor)
    fail("Must specify sampling-rate conversion factor via '-to' or '-by' option");

	inCount=setnote(pp[1],pp[2],output);	/* position output file*/
	setnote(pp[0],pp[2],input);  /* position input file*/

  nChans = sfchans(&sfdesc[input]);
  outCount = ((inCount)*factor)+nChans;

	OBUFFSIZE = IBUFFSIZE*factor+2.0;

	X1 = (HWORD *) calloc(IBUFFSIZE,sizeof(HWORD));
 	X2 = (HWORD *) calloc(IBUFFSIZE,sizeof(HWORD));
	Y1 = (float *) calloc(OBUFFSIZE,sizeof(float));
	Y2 = (float *) calloc(OBUFFSIZE,sizeof(float));
	Xout = (float *) calloc(OBUFFSIZE*nChans,sizeof(float));

	printf("\nStarting Conversion... input samples %d\n", inCount);
	outCountReal = doresample(factor, outCount, nChans,
			interpFilt, linearInterp, largeFilter, filterFile);
	printf("\nFinished Conversion... output samples %d\n", outCountReal);
	endnote(1);
	if (outCountReal <= 0) fail("Bad parameters");

	if (trace && (outCount != outCountReal))
      fprintf(stderr,
	      "outCount = %d, outCountReal = %d\n",outCount,outCountReal);
	
	printf("...Conversion Finished:  %d output samples.\n\n",outCount);
	free(X1);
	free(X2);
	free(Y1);
	free(Y2);
	free(Xout);
	return(0);
}

static int readData(
		    HWORD *outPtr1, 
		    HWORD *outPtr2, 
		    int dataArraySize, 
		    int nChans, int Xoff)
    /* return: 0 - notDone */
    /*        >0 - index of last sample */
{
    int Nsamps, val=0;
    HWORD *dataStart;
//    HWORD *inend = inArray + inCount*nChans;

    dataStart = outPtr1;
    Nsamps = dataArraySize - Xoff; /* Calculate number of samples to get */
    outPtr1 += Xoff;		/* Start at designated sample number */
    if (nChans==1) {
	for (; Nsamps>0; Nsamps--) {
	    if (!GETIN(xin,input)) break;
	    *outPtr1++ = (HWORD)xin[0];
	}
    } else {
		outPtr2 += Xoff;		/* Start at designated sample number */
		for (; Nsamps>0; Nsamps--) {
	    if (!GETIN(xin,input)) break;
	    *outPtr1++ = (HWORD)xin[0];
	    *outPtr2++ = (HWORD)xin[1];
	}
    }
    if (Nsamps > 0) {
	val = outPtr1 - dataStart; /* (Calc return value) */
	while (--Nsamps >= 0) {	/*   fill unread spaces with 0's */
	    *outPtr1++ = 0;	/*   and return FALSE */
	    if (nChans==2)
	      *outPtr2++ = 0;
	}
    }
//	fprintf(stderr,"leaving readdata %d \n",val);
	return(val);
}


#ifdef DEBUG
static int pof = 0;		/* positive overflow count */
static int nof = 0;		/* negative overflow count */
#endif

static INLINE HWORD WordToHword(WORD v, int scl)
{
    HWORD out;
    WORD llsb = (1<<(scl-1));
    v += llsb;		/* round */
    v >>= scl;
    if (v>MAX_HWORD) {
#ifdef DEBUG
	if (pof == 0)
	  fprintf(stderr, "*** libsound: resample: sound sample overflow\n");
	else if ((pof % 10000) == 0)
	  fprintf(stderr, "*** libsound: resample: "
		  "another ten thousand overflows\n");
	pof++;
#endif
	v = MAX_HWORD;
    } else if (v < MIN_HWORD) {
#ifdef DEBUG
	if (nof == 0)
	  fprintf(stderr, "*** resample: sound sample (-) overflow ***\n");
	else if ((nof % 1000) == 0)
	  fprintf(stderr, "*** resample: another thousand (-) overflows **\n");
	nof++;
#endif
	v = MIN_HWORD;
    }	
    out = (HWORD) v;
    return out;
}

/* Sampling rate conversion using linear interpolation for maximum speed.
 */
static int 
  SrcLinear(HWORD X[], float Y[], double factor, UWORD *Time, UHWORD Nx)
{
	HWORD iconst;
	HWORD *Xp;
	float *Ystart;
	WORD v,x1,x2;
    
	double dt;                  /* Step through input signal */ 
	UWORD dtb;                  /* Fixed-point version of Dt */
	UWORD endTime;              /* When Time reaches EndTime, return to user */
    
	dt = 1.0/factor;            /* Output sampling period */
	dtb = dt*(1<<Np) + 0.5;     /* Fixed-point representation */
    
	Ystart = Y;
	endTime = *Time + (1<<Np)*(WORD)Nx;
	while (*Time < endTime){
		iconst = (*Time) & Pmask;
		Xp = &X[(*Time)>>Np];      /* Ptr to current input sample */
		x1 = *Xp++;
		x2 = *Xp;
		x1 *= ((1<<Np)-iconst);
		x2 *= iconst;
		v = x1 + x2;
		*Y++ = WordToHword(v,Np);   /* Deposit output */
		*Time += dtb;		    /* Move to next sample by time increment */
	}
	return (Y - Ystart);            /* Return number of output samples */
}

static int resampleFast(  /* number of output samples returned */
    double factor,		/* factor = Sndout/Sndin */
    int outCount,		/* number of output samples to compute */
    int nChans)			/* number of sound channels (1 or 2) */
{
	UWORD Time, Time2;		/* Current time/pos in input sample */
	UHWORD Xp, Ncreep, Xoff, Xread;
	int OBUFFSIZE = (int)(((double)IBUFFSIZE)*factor+2.0);
	UHWORD Nout, Nx;
	int i, outcnt=0, Ycount, last;
	float *startXout = Xout;

	printf("in fast mode\n");
  
	Xoff = 10;

	Nx = IBUFFSIZE - 2*Xoff;     /* # of samples to process each iteration */
	last = 0;			/* Have not read last input sample yet */
	Ycount = 0;			/* Current sample and length of output file */

	Xp = Xoff;			/* Current "now"-sample pointer for input */
	Xread = Xoff;		/* Position in input array to read into */
	Time = (Xoff<<Np);		/* Current-time pointer for converter */
    
	for (i=0; i<Xoff; X1[i++]=0); /* Need Xoff zeros at begining of sample */
	for (i=0; i<Xoff; X2[i++]=0); /* Need Xoff zeros at begining of sample */

	do {
		if (!last){		/* If haven't read last sample yet */
	    last = readData(X1, X2, IBUFFSIZE, nChans, (int)Xread);
	printf("after readData\n");
	    if (last && (last-Xoff<Nx)) { /* If last sample has been read... */
				Nx = last-Xoff;	/* ...calc last sample affected by filter */
				if (Nx <= 0) break;
	    }
		}

	/* Resample stuff in input buffer */
	Time2 = Time;
	Nout=SrcLinear(X1,Y1,factor,&Time,Nx);
	if (nChans==2) Nout=SrcLinear(X2,Y2,factor,&Time2,Nx);

	Time -= (Nx<<Np);	/* Move converter Nx samples back in time */
	Xp += Nx;		/* Advance by number of samples processed */
	Ncreep = (Time>>Np) - Xoff; /* Calc time accumulation in Time */
	if (Ncreep) {
		Time -= (Ncreep<<Np);    /* Remove time accumulation */
		Xp += Ncreep;            /* and add it to read pointer */
	}
	for (i=0; i<IBUFFSIZE-Xp+Xoff; i++) { /* Copy part of input signal */
		X1[i] = X1[i+Xp-Xoff]; /* that must be re-used */
		if (nChans==2)
			X2[i] = X2[i+Xp-Xoff]; /* that must be re-used */
	}
	if (last) {		/* If near end of sample... */
		last -= Xp;		/* ...keep track were it ends */
		if (!last)		/* Lengthen input by 1 sample if... */
			last++;		/* ...needed to keep flag TRUE */
	}
	Xread = i;		/* Pos in input buff to read new data into */
	Xp = Xoff;
	
	Ycount += Nout;
	if (Ycount>outCount) {
	    Nout -= (Ycount-outCount);
	    Ycount = outCount;
	}

	if (Nout > OBUFFSIZE) /* Check to see if output buff overflowed */
	  return err_ret("Output array overflow");
		outcnt=Nout;
	if (nChans==1) {
			BADDOUT(Y1,1, outcnt);
	} else {
	    register float *y1p = &(Y1[0]);
	    register float *y2p = &(Y2[0]);

	    while (Nout--) {
				*Xout++ = *y1p++; // have to merge chans
				*Xout++ = *y2p++;
			}
			BADDOUT(startXout,1,outcnt);
			Xout =startXout;
	}
    } while (Ycount<outCount); /* Continue until done */
	printf("leaving fast mode\n");
   return(Ycount);		/* Return # of samples in output file */
}

static int resampleWithFilter(  /* number of output sample returned */
	double factor,		/* factor = Sndout/Sndin */
	int outCount,		/* number of output samples to compute */
	int nChans,			/* number of sound channels (1 or 2) */
	BOOL interpFilt,		/* TRUE means interpolate filter coeffs */
	HWORD Imp[], HWORD ImpD[],
	UHWORD LpScl, UHWORD Nmult, UHWORD Nwing)
{
	UWORD Time, Time2;		/* Current time/pos in input sample */
	UHWORD Xp, Ncreep, Xoff, Xread;
	int OBUFFSIZE = (int)(((double)IBUFFSIZE)*factor+2.0);
	UHWORD Nout, Nx;
	int i, outcnt=0, Ycount, last;
	float *startXout = Xout;

 	printf("in resampleWithFilter\n");

	/* Account for increased filter gain when using factors less than 1 */
	if (factor < 1)
		LpScl = LpScl*factor + 0.5;
	/* Calc reach of LP filter wing & give some creeping room */
	Xoff = ((Nmult+1)/2.0) * MAX(1.0,1.0/factor) + 10;
	if (IBUFFSIZE < 2*Xoff)      /* Check input buffer size */
		return err_ret("IBUFFSIZE (or factor) is too small");
	Nx = IBUFFSIZE - 2*Xoff;     /* # of samples to process each iteration */
    
	last = 0;			/* Have not read last input sample yet */
	Ycount = 0;			/* Current sample and length of output file */
	Xp = Xoff;			/* Current "now"-sample pointer for input */
	Xread = Xoff;		/* Position in input array to read into */
	Time = (Xoff<<Np);		/* Current-time pointer for converter */
    
	for (i=0; i<Xoff; X1[i++]=0); /* Need Xoff zeros at begining of sample */
	for (i=0; i<Xoff; X2[i++]=0); /* Need Xoff zeros at begining of sample */

	do {
		if (!last){		/* If haven't read last sample yet */
	    last = readData(X1, X2, IBUFFSIZE, nChans, (int)Xread);
			if (last && (last-Xoff<Nx)) { /* If last sample has been read... */
				Nx = last-Xoff;	/* ...calc last sample affected by filter */
				if (Nx <= 0)
				break;
			}
		}
		/* Resample stuff in input buffer */
		Time2 = Time;
		if (factor >= 1) {	/* SrcUp() is faster if we can use it */
//		fprintf(stderr,"doin' it the fast way\n");
			Nout=SrcUp(X1,Y1,factor,&Time,Nx,Nwing,LpScl,Imp,ImpD,interpFilt);
			if (nChans==2)
				Nout=SrcUp(X2,Y2,factor,&Time2,Nx,Nwing,LpScl,Imp,ImpD,interpFilt);
		}
		else {
//		fprintf(stderr,"doin' it the slow way\n");
			Nout=SrcUD(X1,Y1,factor,&Time,Nx,Nwing,LpScl,Imp,ImpD,interpFilt);
			if (nChans==2)
				Nout=SrcUD(X2,Y2,factor,&Time2,Nx,Nwing,LpScl,Imp,ImpD,interpFilt);
		}

		Time -= (Nx<<Np);	/* Move converter Nx samples back in time */
		Xp += Nx;		/* Advance by number of samples processed */
		Ncreep = (Time>>Np) - Xoff; /* Calc time accumulation in Time */
		if (Ncreep) {
			Time -= (Ncreep<<Np);    /* Remove time accumulation */
			Xp += Ncreep;            /* and add it to read pointer */
		}
		for (i=0; i<IBUFFSIZE-Xp+Xoff; i++) { /* Copy part of input signal */
			X1[i] = X1[i+Xp-Xoff]; /* that must be re-used */
			if (nChans==2)
				X2[i] = X2[i+Xp-Xoff]; /* that must be re-used */
		}
		if (last) {		/* If near end of sample... */
			last -= Xp;		/* ...keep track were it ends */
			if (!last)		/* Lengthen input by 1 sample if... */
				last++;		/* ...needed to keep flag TRUE */
		}
		Xread = i;		/* Pos in input buff to read new data into */
		Xp = Xoff;
	
		Ycount += Nout;
		if (Ycount>outCount) {
			Nout -= (Ycount-outCount);
			Ycount = outCount;
		}

		if (Nout > OBUFFSIZE) /* Check to see if output buff overflowed */
		  return err_ret("Output array overflow");
		outcnt=Nout;
		if (nChans==1) {
			BADDOUT(Y1,1,outcnt);
		}
		else {
			register float *y1p = &(Y1[0]);
			register float *y2p = &(Y2[0]);
			while (Nout--) {
				*Xout++ = *y1p++; // have to merge chans
				*Xout++ = *y2p++;
			}
			BADDOUT(startXout,1,outcnt*2);
			Xout =startXout;
		}
//	printf("Ycount %d outCount %d\n",Ycount,outCount);
	} while (Ycount<outCount); /* Continue until done */
	printf("leaving high quality mode\n");
	return(Ycount);		/* Return # of samples in output file */
}

int doresample(			/* number of output sample returned */
	double factor,		/* factor = Sndout/Sndin */
	int outCount,		/* number of output samples to compute */
	int nChans,			/* number of sound channels (1 or 2) */
	BOOL interpFilt,		/* TRUE means interpolate filter coeffs */
	int fastMode,			/* 0 = highest quality, slowest speed */
	BOOL largeFilter,		/* TRUE means use 65-tap FIR filter */
	char *filterFile)		/* NULL for internal filter, else filename */
{
	UHWORD LpScl;		/* Unity-gain scale factor */
	UHWORD Nwing;		/* Filter table size */
	UHWORD Nmult;		/* Filter length for up-conversions */
	HWORD *Imp=0;		/* Filter coefficients */
	HWORD *ImpD=0;		/* ImpD[n] = Impp[n+1]-Impp[n] */

	printf("in doresample\n");
  
	if (fastMode)
		return resampleFast(factor,outCount,nChans);

#ifdef DEBUG
    /* Check for illegal constants */
    if (Np >= 16)
      return err_ret("Error: Np>=16");
    if (Nb+Nhg+NLpScl >= 32)
      return err_ret("Error: Nb+Nhg+NLpScl>=32");
    if (Nh+Nb > 32)
      return err_ret("Error: Nh+Nb>32");
#endif
    
	/* Set defaults */

	if (filterFile != NULL && *filterFile != '\0') {
		if (readFilter(filterFile, &Imp, &ImpD, &LpScl, &Nmult, &Nwing))
	 	 return err_ret("could not find filter file, "
	       "or syntax error in contents of filter file");
    } else if (largeFilter) {
	Nmult = LARGE_FILTER_NMULT;
	Imp = LARGE_FILTER_IMP;	/* Impulse response */
	ImpD = LARGE_FILTER_IMPD;	/* Impulse response deltas */
	LpScl = LARGE_FILTER_SCALE;	/* Unity-gain scale factor */
	Nwing = LARGE_FILTER_NWING;	/* Filter table length */
    } else {
	Nmult = SMALL_FILTER_NMULT;
	Imp = SMALL_FILTER_IMP;	/* Impulse response */
	ImpD = SMALL_FILTER_IMPD;	/* Impulse response deltas */
	LpScl = SMALL_FILTER_SCALE;	/* Unity-gain scale factor */
	Nwing = SMALL_FILTER_NWING;	/* Filter table length */
    }
#if DEBUG
    fprintf(stderr,"Attenuating resampler scale factor by 0.95 "
	    "to reduce probability of clipping\n");
#endif
    LpScl *= 0.95;
	return resampleWithFilter(factor,outCount,nChans, 
			      interpFilt, Imp, ImpD, LpScl, Nmult, Nwing);
}

/* Sampling rate up-conversion only subroutine;
 * Slightly faster than down-conversion;
 */
static int SrcUp(HWORD X[], float Y[], double factor, UWORD *Time,
		 UHWORD Nx, UHWORD Nwing, UHWORD LpScl,
		 HWORD Imp[], HWORD ImpD[], BOOL Interp)
{
	HWORD *Xp;
	float *Ystart;
	WORD v;
    
	double dt;                  /* Step through input signal */ 
	UWORD dtb;                  /* Fixed-point version of Dt */
	UWORD endTime;              /* When Time reaches EndTime, return to user */
    
	dt = 1.0/factor;            /* Output sampling period */
	dtb = dt*(1<<Np) + 0.5;     /* Fixed-point representation */
    
	Ystart = Y;
	endTime = *Time + (1<<Np)*(WORD)Nx;
	while (*Time < endTime){
		Xp = &X[*Time>>Np];      /* Ptr to current input sample */
		/* Perform left-wing inner product */
		v = FilterUp(Imp, ImpD, Nwing, Interp, Xp, (HWORD)(*Time&Pmask),-1);
		/* Perform right-wing inner product */
		v += FilterUp(Imp, ImpD, Nwing, Interp, Xp+1, 
		      (HWORD)((-*Time)&Pmask),1);
		v >>= Nhg;		/* Make guard bits */
		v *= LpScl;		/* Normalize for unity filter gain */
		*Y++ = WordToHword(v,NLpScl);   /* strip guard bits, deposit output */
		*Time += dtb;		/* Move to next sample by time increment */
	}
	return (Y - Ystart);        /* Return the number of output samples */
}


/* Sampling rate conversion subroutine */

static int SrcUD(HWORD X[], float Y[], double factor, UWORD *Time,
		 UHWORD Nx, UHWORD Nwing, UHWORD LpScl,
		 HWORD Imp[], HWORD ImpD[], BOOL Interp)
{
	HWORD *Xp;
	float *Ystart;
	WORD v;
    
	double dh;                  /* Step through filter impulse response */
	double dt;                  /* Step through input signal */
	UWORD endTime;              /* When Time reaches EndTime, return to user */
	UWORD dhb, dtb;             /* Fixed-point versions of Dh,Dt */
    
	dt = 1.0/factor;            /* Output sampling period */
	dtb = dt*(1<<Np) + 0.5;     /* Fixed-point representation */
    
	dh = MIN(Npc, factor*Npc);  /* Filter sampling period */
	dhb = dh*(1<<Na) + 0.5;     /* Fixed-point representation */
    
	Ystart = Y;
	endTime = *Time + (1<<Np)*(WORD)Nx;
	while (*Time < endTime){
		Xp = &X[*Time>>Np];	/* Ptr to current input sample */
		v = FilterUD(Imp, ImpD, Nwing, Interp, Xp, (HWORD)(*Time&Pmask),
		     -1, dhb);	/* Perform left-wing inner product */
		v += FilterUD(Imp, ImpD, Nwing, Interp, Xp+1, (HWORD)((-*Time)&Pmask),
		      1, dhb);	/* Perform right-wing inner product */
		v >>= Nhg;		/* Make guard bits */
		v *= LpScl;		/* Normalize for unity filter gain */
		*Y++ = WordToHword(v,NLpScl);   /* strip guard bits, deposit output */
		*Time += dtb;		/* Move to next sample by time increment */
	}
	return (Y - Ystart);        /* Return the number of output samples */
}

static int err_ret(char *s)
{
    fprintf(stderr,"resample: %s \n\n",s); /* Display error message  */
    return -1;
}

void profile(void)
{
	UG_INTRO("resample",resample)
}
