#define MAXTERMS 33
#define MACH1 1080.0
#define IWRAP(x,y) (((x)>=(y))?(x)-(y):(((x)<0)?(x)+(y):(x)))
#define FABS(x) (((x)<0.0)?-(x):(x))
#define SIZE 1024

#include <stdio.h>
#include <math.h>
#include "ugens.h"
#include "sfheader.h"
#include "nobug.h"

extern SFHEADER sfdesc[NFILES];
float array[SIZE],tabs[2];	/* for lineset */
int lineset, space_called;

/* This is the main routine that is called by cmix. Its arguments are as
   follows:

   place (inskip, outskip, dur, amp, rho [x], theta [y], recdist, rvbamp)

   The first 3 are as in mix.  "amp" is the amplitude factor for the input
   sig. If "recdist" is positive, location of source is entered in polar
   coordinates: rho (feet), theta (degrees from front, clockwise). If "rec-
   dist" is neg., loc. entered in cartesian: x, y (in feet, listener at 0,0).
   "recdist" is distance in feet between the "mikes" or "ears". "Rvbamp" is fac.
   for reverberation, between 0 and 1.
									*/
extern float rvb_time;         /* From space.c, for time limit. */
extern int tapsize;            /* For allocating space for tap delay */

extern	closesf(void);
extern	setnote(double, double, int);
extern tableset(float, int, float *);   /* set up table for amp */
extern rvb_reset(double *);        /* resets reverb & tap delay */
extern	roomtrig (double,double,double,double[2][13],double [2][13],int);
extern long tap_set(double[2][13], long[2][13], int);
extern	airfil_set (double[2][13], int, double [2][13][3]);
extern	earfil_set (double[2][13], int, double [2][13][33], double Firtaps[2][13][34]);
extern	mike_set(double[2][13], double[2][13]);
extern	get_tap (int, long [2][13], double[2][13], double *); /* get delayed samps */
extern	air (double[2][13], double [2][13][3]);                  /* air absorpt. filters */
extern	wall (double[2][13]);                         /* wall absorpt. filters */
extern	ear (double[2][13], long, double[2][13][33], double[2][13][34]); /* binaural angle filters */
extern	RVB (double *, double *, int);			/* run through reverberator */
extern	mike(double[2][13], double[2][13]);			/* do cartioid mike effect */
extern	endnote(int);
extern	cfree(char *);

double place (float *,int, double *);

int UseMikes = 0;
double wrap();
double Airdata[2][13][3]; 
double  Rho[2][13], Theta[2][13], Sig[2][13], MikeAmps[2][13];
double Firtaps[2][13][MAXTERMS+1], Fircoeffs[2][13][MAXTERMS];   
long outlocs[2][13];

double 
place (float *p,int n_args, double *pp)
{
	double dist, R, T, dur;
	int nsamps, tapcount, N, i, Nend;
	int flag, ear_flag, intap, cartflag=0, z, j,chn;
	double rvbsig[2], roomsig[2];
	float input[4], outbox[4], rvbamp, amp;
	double *tapdel;

	p=p;
	if(n_args > 9)
	{
		fprintf(stderr,"Wrong number of args for move!\n");
		closesf();
		return 0;
	}
	if(n_args < 8)
	{
		fprintf(stderr,"Wrong number of args for move!\n");
		closesf();
		return 0;
	}
	if(n_args == 8)
	{
		fprintf(stderr,"Using Channel 0 as input by default.\n");
		chn = 0;
	}
	if(n_args == 9)
	{
		if(pp[8] == 0) {
			fprintf(stderr,"Using Channel 0 as input.\n");
			chn = 0;}
		else if(pp[8] == 1) {
			fprintf(stderr,"Using Channel 1 as input.\n");
			chn = 1;}
		else {
			fprintf(stderr,"Using Channel 0 as input.\n");
			chn = 0;}
	}
	dur = pp[2]; 
	dur = ((dur > 0.0) ? dur - pp[0] : -dur); 
	setnote(pp[0], dur, 0);           /* set input file */
	nsamps = setnote(pp[1], dur, 1);  /* set output file */
	dist = pp[6];
	if (dist < 0) {
		cartflag = 1; 	/* cartesian coordinates */
		dist = -dist;
		fprintf(stderr, "Using cartesian coordinate system.\n");
	}
//	if(sfchans(&sfdesc[0]) != 1) {
//		fprintf(stderr,"Input file must be 1-channel.\n");
//		return;
//	}
	if(sfchans(&sfdesc[1]) != 2) {   
		fprintf(stderr,"Output file must be 2-channel.\n");
		return 0;
	}
	if(!space_called) {
		fprintf(stderr,"You must call setup routine 'space' first!\n");
		return 0;
	}

	/* allocate space for tap delay based on room dimensions */

	if((tapdel = (double *) calloc((unsigned)(tapsize+8), sizeof(double))) == NULL)
	{
		fprintf(stderr, "Not enough memory for tap delay!\n");
		return 0;
	}

	/* flag for use of ear filters */

	ear_flag = (!UseMikes && dist < 0.8 && dist != 0.0);
	if(ear_flag)
		printf("Running in binaural mode.\n");

	z = SR/100;		      /* the control rate for amp curve */
	if(!lineset) {
		for(i=0; i<SIZE; i++) array[i] = 1;
		fprintf(stderr, "Setting phrase curve to all 1's.\n");
		lineset = 1;
	}
	j = 0;
	tableset((float)dur, SIZE, tabs);   /* set up table for amp */

	rvbamp = (float)pp[7];             /* amp factor for reverberated sig */
	flag = 1; 															/* reset all filters, etc. */
	rvb_reset(tapdel);											/* resets reverb & tap delay */
	intap = 0;															/* the starting element in tap delay */

	/* set up room */
	R = pp[4];
	T = pp[5];
	if(roomtrig (R,T,dist,Rho,Theta,cartflag)) return 0; 

	/* set taps, return max samp */
	tapcount = tap_set(Rho, outlocs, ear_flag);
	airfil_set (Rho, flag, Airdata);
	if (ear_flag)
		earfil_set (Theta, flag, Fircoeffs, Firtaps);   
	else
		mike_set(Theta, MikeAmps);

	/* the processing loop for the input signal */
	printf("processing input signal...");
	
	for (N = 0; N < nsamps; N++) 
	{	
		if(!GETIN(input,0)) break;      /* get input samp until EOF */ 
		if(intap>=tapsize) intap -= tapsize;   /* input tap location */
		while (!j--){
			amp = tablei(N,array,tabs) * pp[3];
			j = z;
		}
//		tapdel[intap] = input[0]*amp;          /* add to delay */
		if(chn) { tapdel[intap] = input[1]*amp;} // add right chan to delay
		else {tapdel[intap] = input[0]*amp; }   // add left chan to rdelay
		get_tap (intap, outlocs, Sig, tapdel); /* get delayed samps */
		air (Sig, Airdata);                  /* air absorpt. filters */
		wall (Sig);                         /* wall absorpt. filters */
		if (ear_flag)
		  ear (Sig, N, Fircoeffs, Firtaps); /* binaural angle filters */
		roomsig[0] = roomsig[1] = 0.0; 
		for (i = 1; i < 13; ++i) {			/* sum reflected signals */
			roomsig[0] += Sig[0][i];		/* as input for RVB      */
			roomsig[1] += Sig[1][i]; 
		}
		/* scale reverb input by amp factor */
		roomsig[0] *= rvbamp;  roomsig[1] *= rvbamp; 
		RVB (roomsig, rvbsig, N);			/* run through reverberator */

		if (!ear_flag) {
			roomsig[0] = roomsig[1] = 0.0;	/* zero these again */
			mike(Sig, MikeAmps);			/* do cartioid mike effect */
			for (i = 1; i < 13; ++i) {		/* sum reflected signals */
				roomsig[0] += Sig[0][i];
				roomsig[1] += Sig[1][i]; 
			}
		}

		/* sum the direct, early refl. & reverbed sigs  */
		
//	printf("Sig[0][0] %f roomsig[0] %f rvbsig[0] %f\n",Sig[0][0],roomsig[0],rvbsig[0]);
//		outbox[0] = Sig[0][0]+roomsig[0]+rvbsig[0];
//		outbox[1] = Sig[1][0]+roomsig[1]+rvbsig[1];
		outbox[0] = Sig[0][0]+roomsig[0];
		outbox[1] = Sig[1][0]+roomsig[1];
		ADDOUT(outbox,1);  
		flag = 0;                     /* turn off reset flag */
		intap++;
	}

	/* the processing loop for the tap delay signal */
	printf("flushing delay lines...");

	nsamps = (N < nsamps) ? N : nsamps;	/* if EOF reached above */
	Nend = nsamps + tapcount;
	for (N = nsamps; N < Nend; N++) 
	{	
		if(intap>=tapsize) intap -= tapsize;              
		tapdel[intap] = 0.0;           
//		if(chn) { tapdel[intap] = input[1]*amp;} // add right chan to delay
//		else {tapdel[intap] = input[0]*amp; }   // add left chan to rdelay
		get_tap (intap, outlocs, Sig, tapdel); 
		air (Sig, Airdata);   
		wall (Sig);    
		if (ear_flag)
		   ear (Sig, N, Fircoeffs, Firtaps); 
		roomsig[0] = roomsig[1] = 0.0;
		for (i = 1; i < 13; ++i) {
		   roomsig[0] += Sig[0][i];
		   roomsig[1] += Sig[1][i];    
		}
		/* scale by amp factor */
		roomsig[0] *= rvbamp;  roomsig[1] *= rvbamp; 
		RVB (roomsig, rvbsig, N);    

		if (!ear_flag) {
			roomsig[0] = roomsig[1] = 0.0;	/* zero these again */
			mike(Sig, MikeAmps);			/* do cartioid mike effect */
			for (i = 1; i < 13; ++i) {		/* sum reflected signals */
				roomsig[0] += Sig[0][i];
				roomsig[1] += Sig[1][i]; 
			}
		}
		outbox[0] = Sig[0][0] + roomsig[0] + rvbsig[0];
		outbox[1] = Sig[1][0] + roomsig[1] + rvbsig[1];
		ADDOUT(outbox,1);  
		intap++;
	}
	/* clean out the reverb unit */
	printf("flushing reverb...");
	
	i = 0;
	N = Nend;  
	Nend = N + (rvb_time + 0.25) * SR;
	while (N < Nend) {
		roomsig[0] = roomsig[1] = 0.0;
		RVB(roomsig, rvbsig, N);
		outbox[0] = rvbsig[0];
		outbox[1] = rvbsig[1];
		ADDOUT(outbox,1);
		i++;
		N++;
	}
	printf("\n");
	endnote(1);

	cfree((char *)tapdel);
	return 0;
}
