/* Plucked string instrument using strum.
    Use with fret.c and bend.c to play single pluck sounds.

    start(start time, dur, pitch, tf0, tNy, amp, squish)

    tf0 is the decay time at the fundamental frequency of the note.
    tNy is high freq decay time.
    squish is an integer 0 or greater that specifies how 'squishy' the
      plucking implement is.  0 is a hard implement.  5 is more like a finger.

    tNy is normally a lot smaller than tf0.

    Normally the random generator is not started with the same seed for each
    note.  Thus a note called with identical parameters will sound slightly
    different each time it it played, unless the random generator is reseeded
    with the same seed before each note.  This can be done from inside Minc 
    with a call to srand(seed).

    Decay times need not equal the duration of the note.  Short decay times   
    with small squish sound like plucked violins, etc.  Long decay times with
    larger squish sound more like electric guitars, basses, etc.

    Use fret to change some parameter of the sound after it is started with
    start.  For example, a note with a long duration will end abruptly unless
    you call fret right before it stops with a very short decay time.

    Use bend to gliss or add vibrato, or whatever.  If you want to start 
    the gliss immediately, just call start with zero duration.   */

#include "../H/ugens.h"
#include "../H/sfheader.h"
#include "../macros/macros.h"
#include "strums.h"

double start(float *p,int n_args, double *pp);
extern int setnote (float, float, int); 
extern void sset(float ,float ,float,strumq *);
extern endnote(int);
extern void randfill(float ,int,strumq *);
extern float strum(float,strumq *);

strumq strumq1;

double start(float *p,int n_args, double *pp)
{
  float freq,out[4], start, fdur;
  int squish;
  long i,nsamps;
//  extern strumq strumq1;

  out[0]=out[1]=out[2]=out[3]=0.0;
	start = (float) pp[0];
	fdur = (float) pp[1];

  nsamps=setnote(start,fdur,0);

  freq=cpspch(pp[2]);
  squish = pp[6];
   
  sset(freq,pp[3],pp[4],&strumq1);
  randfill((float)pp[5],squish,&strumq1);

  for (i=0;i<nsamps;i++) {
      out[0]=strum(0.,&strumq1);
//   printf("val %f\n",out[0]);
      ADDOUT(out,0);
  }
  endnote(0);
  return 0.0;
}
