#include "../H/sfheader.h"
#include "../H/ugens.h"
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <bsd_mem.h>
#include <errno.h>

#define L_SET 0

static SFCODE	ampcode = {
	SF_MAXAMP,
	sizeof(SFMAXAMP) + sizeof(SFCODE)
}; 

int readIrcamHeader(SFHEADER *);
extern int putsfcode(SFHEADER *,void *,SFCODE *);
extern int getsfcode(SFHEADER *, int);
int writeHeader(int sf,SFHEADER  *header);	/* writes out whatever type of header was read in */
int getheadersize(SFHEADER *header);
int putlength(char *sfname,int sf,SFHEADER *header);
int readHeader(int sf,SFHEADER *header);

int getheadersize(SFHEADER *header)
{
int offset;
#if defined(NeXT) || defined(NEXT)
	if(nsfmagic(header) == SND_MAGIC) {
		/* native NeXT files may have arbitrary offset size */
		offset = (header)->sfinfo.NeXTheader.dataLocation;
		/* make sure we dont return bad or negative value! */
		return (offset >= SIZEOF_NeXT_HEADER) ? offset : SIZEOF_NeXT_HEADER;
	}
	else return(SIZEOF_BSD_HEADER);
#else
	return(SIZEOF_BSD_HEADER);
#endif
}

#if defined(NeXT) || defined(NEXT)

//int putlength(char *sfname,FILE * sf,SFHEADER *header)
int putlength(char *sfname,int sf,SFHEADER *header)
{
struct stat st;
	int headersize = 0;

	if(stat(sfname,&st))  {
		fprintf(stderr, "putlength:  Couldn't stat file\n");
		return(-1);
	}
	/* dont do anything if file is Ircam style */
	if((header)->sfinfo.NeXTheader.magic != SND_MAGIC){
		return 1;
		}
	if((header)->sfinfo.sf_magic == 0) headersize = SIZEOF_NeXT_HEADER;
	else headersize = SIZEOF_BSD_HEADER;
	(header)->sfinfo.NeXTheader.dataSize = (int)st.st_size - headersize;

	if(lseek(sf,0,SEEK_SET) < 0) {
		fprintf(stderr, "putlength: Bad lseek on file: %s\n", sfname);
		return(-1);
	}
	if(write(sf,header,headersize) != headersize) {
//	if(fwrite(header,headersize,1,sf) == NULL) {
		fprintf(stderr, "putlength:  Bad header write on file\n");
		return(-1);
	}
	return (0);
}

//readHeader(FILE * sf,SFHEADER *header)
readHeader(int sf,SFHEADER *header)
{
 int i;
	SFMAXAMP sfmnew;
	char *pointer;
/* read in 1kb first, to see if hybrid header is present */

	if((i=read(sf,(char *) header,SIZEOF_BSD_HEADER)) != SIZEOF_BSD_HEADER)
//        if((i=fread((char *) header,SIZEOF_BSD_HEADER,1,sf)) != 1)
	if((i=read(sf,(char *)header,SIZEOF_BSD_HEADER)) != 1) {
			return 1;}
/* if file has IRCAM-style header */
	if((header)->sfinfo.NeXTheader.magic == SF_MAGIC)
		return(readIrcamHeader(header));
/* if magic number is neither IRCAM nor local, we cant read it */
/* in case it is byte-swapped, give a message saying so */
	if((header)->sfinfo.NeXTheader.magic == 1688404224) {
		fprintf(stderr, "This file must be byte-swapped before using!\n");
		return 1;
	}
	if((header)->sfinfo.NeXTheader.magic != SND_MAGIC)
	return 1;
/* if hybrid header is there, we are done here */
	if((header)->sfinfo.sf_magic == SF_MAGIC) {
		return 0;}

/* else, if file being read is native NeXT soundfile... */

/* zero out new header after NeXT portion */
	pointer = (char *)header;
	for(i=29; i<SIZEOF_BSD_HEADER; i++) *(pointer+i) = 0;
/* set file pointer to beginning of sound data */
//        if(lseek(sf, (long) header->sfinfo.NeXTheader.dataLocation, L_SET) < 0) {
		if(lseek(sf, (off_t)sizeof(header->sfinfo.NeXTheader.dataLocation), L_SET) < 0) {
//        if(fseek(sf, (off_t)sizeof(header->sfinfo.NeXTheader.dataLocation), SEEK_SET) < 0) {
			fprintf(stderr, "Unable to reset file pointer.\n");
    	return 1;
    }
        /* load all values from native header into hybrid header */
        (header)->sfinfo.sf_magic = 0; /* so we know it is native NeXT file */
        (header)->sfinfo.sf_srate = (header)->sfinfo.NeXTheader.samplingRate;
        (header)->sfinfo.sf_chans = (header)->sfinfo.NeXTheader.channelCount;
        switch(header->sfinfo.NeXTheader.dataFormat) {
        case SND_FORMAT_FLOAT:
                (header)->sfinfo.sf_packmode = SF_FLOAT;
                break;
        case SND_FORMAT_LINEAR_16:
                (header)->sfinfo.sf_packmode = SF_SHORT;
                break;
        default:
                fprintf(stderr, "Cmix can only read float and short files.\n");
                return 1;
        }
	for(i=0; i<2 /*(header)->sfinfo.sf_chans*/; i++) {
			sfmaxamp(&sfmnew,i)=0;
			sfmaxamploc(&sfmnew,i)=0;
		}
		sfmaxamptime(&sfmnew) = 0;
		putsfcode(header,&sfmnew,&ampcode);
   return 0;
}

readIrcamHeader(SFHEADER *hd)
{
char *ptr = (char *) hd;
int i;
	SFMAXAMP sfmnew;
/* copy everything in header (minus a small chunk) to the location
   in the header it needs to be in a hybrid header.  We loose 28 chars
   from the end of the comment, but who uses that much comment?  */

	bcopy(ptr, ptr + SIZEOF_NeXT_HEADER, sizeof(SFHEADER) - SIZEOF_NeXT_HEADER);
//memmove(ptr + SIZEOF_NeXT_HEADER, ptr,sizeof(SFHEADER) - SIZEOF_NeXT_HEADER);
/* and then zero out beginning and set the new magic number */
 bzero(ptr, SIZEOF_NeXT_HEADER);
/* if no amp code in header, put one */
	if(getsfcode(hd, SF_MAXAMP) == NULL) {
		for(i=0; i<2; i++) {
			sfmaxamp(&sfmnew,i)=0;
			sfmaxamploc(&sfmnew,i)=0;
		}
		sfmaxamptime(&sfmnew) = 0;
		putsfcode(hd,&sfmnew,&ampcode);
	}
	/* set the magic number at the beginning to be the IRCAM magic number,
	   to signal that this should be written back out as such.	*/
        hd->sfinfo.NeXTheader.magic = SF_MAGIC;
        return 0;
}

//writeHeader(FILE * sf,SFHEADER  *header)	/* writes out whatever type of header was read in */
writeHeader(int sf,SFHEADER  *header)	/* writes out whatever type of header was read in */
{
	int headersize, dataloc, bytes;
	static char blank[SIZEOF_NeXT_HEADER];

	switch((header)->sfinfo.NeXTheader.magic) {
	case SND_MAGIC:		/* native or hybrid NeXT soundfile */
		printf("Writing hybrid header\n");
		if((header)->sfinfo.sf_packmode == SF_FLOAT)
			(header)->sfinfo.NeXTheader.dataFormat =
				SND_FORMAT_FLOAT;
		else (header)->sfinfo.NeXTheader.dataFormat =
			SND_FORMAT_LINEAR_16;
		(header)->sfinfo.NeXTheader.samplingRate =
			(header)->sfinfo.sf_srate;
	(header)->sfinfo.NeXTheader.channelCount =
			(header)->sfinfo.sf_chans;
		if((header)->sfinfo.sf_magic == 0) {	/* Native */
			dataloc = SIZEOF_NeXT_HEADER;
			headersize = SIZEOF_NeXT_HEADER; 
		}
		else {					/* Hybrid */
			dataloc = SIZEOF_BSD_HEADER;
			headersize = SIZEOF_BSD_HEADER; //really going to write 1024 byte header
		}
		(header)->sfinfo.NeXTheader.dataLocation = dataloc;
		if((bytes=write(sf,header,headersize)) != headersize)
//		if((bytes=fwrite(header,headersize,1,sf)) != 1)
				return 1;
		else 
				return 0;
	break;
	case SF_MAGIC:		/* IRCAM style soundfile */
		printf("IRCAM header\n");
		/* try to write out 1 Kbyte starting right after NeXT header */
		headersize = sizeof(SFHEADER) - SIZEOF_NeXT_HEADER;
		if((bytes=write(sf,(char *) header + SIZEOF_NeXT_HEADER, headersize)) != headersize)
//	if(bytes = (fwrite(header+SIZEOF_NeXT_HEADER,headersize,1,sf)) != 1)
			return 1;
		/* this fills in the last 28 bytes */
		if((bytes=write(sf, blank, SIZEOF_NeXT_HEADER)) != SIZEOF_NeXT_HEADER)
//	if(bytes = (fwrite(blank,SIZEOF_NeXT_HEADER,1,sf))  != 1)
		return 1;
		else return 0;
	break;
	default:
		fprintf(stderr, "Unknown magic number: %d\n",
			(header)->sfinfo.NeXTheader.magic);
		return 1;
		break;
	}
}
#else

putlength(char *sfname,int sf,SFHEADER *header)
{
	return 0;	/* do nothing */
}

#endif /* NeXT */
